# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from .complexdop import ComplexDop
from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from .multiplexercase import MultiplexerCase
from .multiplexerdefaultcase import MultiplexerDefaultCase
from .multiplexerswitchkey import MultiplexerSwitchKey
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import AtomicOdxType, ParameterValue, odxstr_to_bool
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Multiplexer(ComplexDop):
    """This class represents a Multiplexer (MUX)

    Multiplexers are used to interpret data stream depending on the
    value of a switch-key (similar to switch-case statements in
    programming languages like C or Java).
    """

    byte_position: int
    switch_key: MultiplexerSwitchKey
    default_case: MultiplexerDefaultCase | None = None
    cases: NamedItemList[MultiplexerCase] = field(default_factory=NamedItemList)
    is_visible_raw: bool | None = None

    @property
    def is_visible(self) -> bool:
        return self.is_visible_raw is True

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Multiplexer":
        """Reads a Multiplexer from Diag Layer."""
        kwargs = dataclass_fields_asdict(ComplexDop.from_et(et_element, context))

        byte_position = int(et_element.findtext("BYTE-POSITION", "0"))
        switch_key = MultiplexerSwitchKey.from_et(
            odxrequire(et_element.find("SWITCH-KEY")), context)

        default_case = None
        if (dc_elem := et_element.find("DEFAULT-CASE")) is not None:
            default_case = MultiplexerDefaultCase.from_et(dc_elem, context)

        cases = NamedItemList(
            [MultiplexerCase.from_et(el, context) for el in et_element.iterfind("CASES/CASE")])

        is_visible_raw = odxstr_to_bool(et_element.get("IS-VISIBLE"))

        return Multiplexer(
            byte_position=byte_position,
            switch_key=switch_key,
            default_case=default_case,
            cases=cases,
            is_visible_raw=is_visible_raw,
            **kwargs)

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()

        odxlinks.update(self.switch_key._build_odxlinks())
        if self.default_case is not None:
            odxlinks.update(self.default_case._build_odxlinks())

        return odxlinks

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        self.switch_key._resolve_odxlinks(odxlinks)
        if self.default_case is not None:
            self.default_case._resolve_odxlinks(odxlinks)

        for mux_case in self.cases:
            mux_case._mux_case_resolve_odxlinks(
                odxlinks, key_physical_type=self.switch_key.dop.physical_type.base_data_type)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        self.switch_key._resolve_snrefs(context)
        if self.default_case is not None:
            self.default_case._resolve_snrefs(context)

        for mux_case in self.cases:
            mux_case._resolve_snrefs(context)

    def _get_case_limits(self, case: MultiplexerCase) -> tuple[AtomicOdxType, AtomicOdxType]:
        key_type = self.switch_key.dop.physical_type.base_data_type
        lower_limit = key_type.make_from(case.lower_limit.value)
        upper_limit = key_type.make_from(case.upper_limit.value)
        if not isinstance(lower_limit, type(upper_limit)) and not isinstance(
                upper_limit, type(lower_limit)):
            odxraise("Upper and lower bounds of limits must compareable")
        return lower_limit, upper_limit

    @override
    def encode_into_pdu(self, physical_value: ParameterValue, encode_state: EncodeState) -> None:

        if encode_state.cursor_bit_position != 0:
            raise EncodeError(f"Multiplexer parameters must be aligned, i.e. bit_position=0, but "
                              f"{self.short_name} was passed the bit position "
                              f"{encode_state.cursor_bit_position}")

        orig_origin = encode_state.origin_byte_position
        encode_state.origin_byte_position = encode_state.cursor_byte_position

        if isinstance(physical_value, (list, tuple)) and len(physical_value) == 2:
            case_spec, case_value = physical_value
        elif isinstance(physical_value, dict) and len(physical_value) == 1:
            case_spec, case_value = next(iter(physical_value.items()))
        else:
            raise EncodeError(
                f"Values of multiplexer parameters must be defined as a "
                f"(case_name, content_value) tuple instead of as '{physical_value!r}'")

        mux_case: MultiplexerCase | MultiplexerDefaultCase
        applicable_cases: list[MultiplexerCase | MultiplexerDefaultCase]

        if isinstance(case_spec, str):
            applicable_cases = [x for x in self.cases if x.short_name == case_spec]
            if not applicable_cases and self.default_case:
                applicable_cases.append(self.default_case)
            if len(applicable_cases) == 0:
                raise EncodeError(
                    f"Multiplexer {self.short_name} does not know any case called {case_spec}")

            odxassert(len(applicable_cases) == 1)
            mux_case = applicable_cases[0]
            if isinstance(mux_case, MultiplexerCase):
                key_value, _ = self._get_case_limits(mux_case)
            else:
                key_value = 0
        elif isinstance(case_spec, int):
            applicable_cases = []
            for x in self.cases:
                lower, upper = cast(tuple[int, int], self._get_case_limits(x))
                if lower <= case_spec and case_spec <= upper:
                    applicable_cases.append(x)

            if len(applicable_cases) == 0:
                if self.default_case is None:
                    raise EncodeError(
                        f"Multiplexer {self.short_name} does not know any case called {case_spec}")
                mux_case = self.default_case
                key_value = case_spec
            else:
                mux_case = applicable_cases[0]
                key_value = case_spec
        elif isinstance(case_spec, MultiplexerCase):
            mux_case = case_spec
            key_value, _ = self._get_case_limits(mux_case)
        elif case_spec is None:
            if self.default_case is None:
                raise EncodeError(f"Multiplexer {self.short_name} does not define a default case")
            key_value = 0
        else:
            raise EncodeError(f"Illegal case specification '{case_spec}' for "
                              f"multiplexer {self.short_name}")

        # the byte position of the switch key is relative to
        # the multiplexer's position
        encode_state.cursor_byte_position = encode_state.origin_byte_position + self.switch_key.byte_position
        encode_state.cursor_bit_position = self.switch_key.bit_position or 0
        self.switch_key.dop.encode_into_pdu(physical_value=key_value, encode_state=encode_state)
        encode_state.cursor_bit_position = 0

        if mux_case.structure is not None:
            # the byte position of the content is specified by the
            # BYTE-POSITION attribute of the multiplexer
            encode_state.cursor_byte_position = encode_state.origin_byte_position + self.byte_position
            mux_case.structure.encode_into_pdu(physical_value=case_value, encode_state=encode_state)

        encode_state.origin_byte_position = orig_origin

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        orig_origin = decode_state.origin_byte_position
        decode_state.origin_byte_position = decode_state.cursor_byte_position

        # Decode the switch key. Its BYTE-POSITION is relative to the
        # that of the multiplexer.
        if self.switch_key.byte_position is not None:
            decode_state.cursor_byte_position = decode_state.origin_byte_position + self.switch_key.byte_position
        decode_state.cursor_bit_position = self.switch_key.bit_position or 0
        key_value = self.switch_key.dop.decode_from_pdu(decode_state)
        decode_state.cursor_bit_position = 0

        if not isinstance(key_value, int):
            odxraise(f"Multiplexer keys must be integers (is '{type(key_value).__name__}'"
                     f" for multiplexer '{self.short_name}')")

        # "If a matching CASE is found, the referenced STRUCTURE is
        # analyzed at the BYTE-POSITION (child element of MUX)
        # relatively to the byte position of the MUX."
        decode_state.cursor_byte_position = decode_state.origin_byte_position + self.byte_position

        applicable_case: MultiplexerCase | MultiplexerDefaultCase | None = None
        for mux_case in self.cases:
            lower, upper = self._get_case_limits(mux_case)
            if lower <= key_value and key_value <= upper:  # type: ignore[operator]
                applicable_case = mux_case
                break

        if applicable_case is None:
            applicable_case = self.default_case

        if applicable_case is None:
            odxraise(
                f"Cannot find an applicable case for value {key_value} in "
                f"multiplexer {self.short_name}", DecodeError)
            decode_state.origin_byte_position = orig_origin
            return (None, None)

        if applicable_case.structure is not None:
            case_value = applicable_case.structure.decode_from_pdu(decode_state)
        else:
            case_value = {}

        result = (applicable_case.short_name, case_value)

        decode_state.origin_byte_position = orig_origin

        return result



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .diagnostictroublecode import DiagnosticTroubleCode
from .exceptions import odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext

if TYPE_CHECKING:
    from .dtcdop import DtcDop


@dataclass(kw_only=True)
class LinkedDtcDop:
    not_inherited_dtc_snrefs: list[str] = field(default_factory=list)
    dtc_dop_ref: OdxLinkRef

    @property
    def not_inherited_dtcs(self) -> NamedItemList[DiagnosticTroubleCode]:
        return self._not_inherited_dtcs

    @property
    def dtc_dop(self) -> "DtcDop":
        return self._dtc_dop

    @property
    def short_name(self) -> str:
        return self._dtc_dop.short_name

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "LinkedDtcDop":
        not_inherited_dtc_snrefs = [
            odxrequire(el.get("SHORT-NAME"))
            for el in et_element.iterfind("NOT-INHERITED-DTC-SNREFS/"
                                          "NOT-INHERITED-DTC-SNREF")
        ]

        dtc_dop_ref = odxrequire(OdxLinkRef.from_et(et_element.find("DTC-DOP-REF"), context))

        return LinkedDtcDop(
            not_inherited_dtc_snrefs=not_inherited_dtc_snrefs, dtc_dop_ref=dtc_dop_ref)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if TYPE_CHECKING:
            self._dtc_dop = odxlinks.resolve(self.dtc_dop_ref, DtcDop)
        else:
            self._dtc_dop = odxlinks.resolve(self.dtc_dop_ref)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        dtc_dop = self._dtc_dop
        not_inherited_dtcs = [
            resolve_snref(ni_snref, dtc_dop.dtcs, DiagnosticTroubleCode)
            for ni_snref in self.not_inherited_dtc_snrefs
        ]

        self._not_inherited_dtcs = NamedItemList(not_inherited_dtcs)



# SPDX-License-Identifier: MIT
import typing
from typing import runtime_checkable

from .decodestate import DecodeState
from .encodestate import EncodeState
from .odxtypes import ParameterValue


@runtime_checkable
class Codec(typing.Protocol):
    """Any object which can be en- or decoded to be transferred over
    the wire implements this API.
    """

    @property
    def short_name(self) -> str:
        return ""

    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        ...

    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        ...

    def get_static_bit_length(self) -> int | None:
        ...



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .element import IdentifiableElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class TeamMember(IdentifiableElement):
    roles: list[str] = field(default_factory=list)
    department: str | None = None
    address: str | None = None
    zipcode: str | None = None  # the tag for this is "ZIP", but `zip` is a keyword in python
    city: str | None = None
    phone: str | None = None
    fax: str | None = None
    email: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "TeamMember":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        roles = [odxrequire(role_elem.text) for role_elem in et_element.iterfind("ROLES/ROLE")]
        department = et_element.findtext("DEPARTMENT")
        address = et_element.findtext("ADDRESS")
        zipcode = et_element.findtext("ZIP")
        city = et_element.findtext("CITY")
        phone = et_element.findtext("PHONE")
        fax = et_element.findtext("FAX")
        email = et_element.findtext("EMAIL")

        return TeamMember(
            roles=roles,
            department=department,
            address=address,
            zipcode=zipcode,
            city=city,
            phone=phone,
            fax=fax,
            email=email,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, cast
from xml.etree import ElementTree

from .element import NamedElement
from .exceptions import odxraise, odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .unit import Unit
from .unitgroupcategory import UnitGroupCategory
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class UnitGroup(NamedElement):
    """A group of units.

    There are two categories of groups: COUNTRY and EQUIV-UNITS.
    """
    category: UnitGroupCategory
    unit_refs: list[OdxLinkRef] = field(default_factory=list)
    oid: str | None = None

    @property
    def units(self) -> NamedItemList[Unit]:
        return self._units

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "UnitGroup":
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        category_str = odxrequire(et_element.findtext("CATEGORY"))
        try:
            category = UnitGroupCategory(category_str)
        except ValueError:
            category = cast(UnitGroupCategory, None)
            odxraise(f"Encountered unknown unit group category '{category_str}'")

        unit_refs = [
            odxrequire(OdxLinkRef.from_et(el, context))
            for el in et_element.iterfind("UNIT-REFS/UNIT-REF")
        ]
        oid = et_element.get("OID")

        return UnitGroup(category=category, unit_refs=unit_refs, oid=oid, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._units = NamedItemList[Unit]([odxlinks.resolve(ref) for ref in self.unit_refs])

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod
from .linearsegment import LinearSegment


@dataclass(kw_only=True)
class LinearCompuMethod(CompuMethod):
    """A compu method which does linear interpoation

    i.e. internal values are converted to physical ones using the
    function `f(x) = (offset + factor * x)/denominator` where `f(x)`
    is the physical value and `x` is the internal value. In contrast
    to `ScaleLinearCompuMethod`, this compu method only exhibits a
    single segment)

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.3.
    """

    @property
    def segment(self) -> LinearSegment:
        return self._segment

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "LinearCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return LinearCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        odxassert(self.category == CompuCategory.LINEAR,
                  "LinearCompuMethod must exhibit LINEAR category")

        odxassert(self.physical_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])
        odxassert(self.internal_type in [
            DataType.A_FLOAT32,
            DataType.A_FLOAT64,
            DataType.A_INT32,
            DataType.A_UINT32,
        ])

        if self.compu_internal_to_phys is None:
            odxraise("LINEAR compu methods require COMPU-INTERNAL-TO-PHYS")
            return

        compu_scales = self.compu_internal_to_phys.compu_scales

        if len(compu_scales) == 0:
            odxraise("LINEAR compu methods expect at least one compu scale within "
                     "COMPU-INTERNAL-TO-PHYS")
            return cast(None, LinearCompuMethod)
        elif len(compu_scales) > 1:
            odxraise("LINEAR compu methods expect at most one compu scale within "
                     "COMPU-INTERNAL-TO-PHYS")

        scale = compu_scales[0]
        self._segment = LinearSegment.from_compu_scale(
            scale, internal_type=self.internal_type, physical_type=self.physical_type)

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        if not self._segment.internal_applies(internal_value):
            odxraise(f"Cannot decode internal value {internal_value!r}", DecodeError)

        return self._segment.convert_internal_to_physical(internal_value)

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        if not self._segment.physical_applies(physical_value):
            odxraise(f"Cannot decode physical value {physical_value!r}", EncodeError)

        return self._segment.convert_physical_to_internal(physical_value)

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        return self._segment.physical_applies(physical_value)

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        return self._segment.internal_applies(internal_value)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import DecodeError, EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from ..odxtypes import ParameterValue
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType
from .tablekeyparameter import TableKeyParameter

if TYPE_CHECKING:
    from ..table import Table


@dataclass(kw_only=True)
class TableStructParameter(Parameter):
    table_key_ref: OdxLinkRef | None = None
    table_key_snref: str | None = None

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "TABLE-STRUCT"

    @property
    def table(self) -> "Table":
        return self._table_key.table

    @property
    def table_key(self) -> TableKeyParameter:
        return self._table_key

    @property
    @override
    def is_required(self) -> bool:
        return True

    @property
    @override
    def is_settable(self) -> bool:
        return True

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "TableStructParameter":

        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        table_key_ref = OdxLinkRef.from_et(et_element.find("TABLE-KEY-REF"), context)
        table_key_snref = None
        if (table_key_snref_elem := et_element.find("TABLE-KEY-SNREF")) is not None:
            table_key_snref = odxrequire(table_key_snref_elem.get("SHORT-NAME"))

        return TableStructParameter(
            table_key_ref=table_key_ref, table_key_snref=table_key_snref, **kwargs)

    def __post_init__(self) -> None:
        if self.table_key_ref is None and self.table_key_snref is None:
            odxraise("Either table_key_ref or table_key_snref must be defined.")

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return super()._build_odxlinks()

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        if self.table_key_ref is not None:
            self._table_key = odxlinks.resolve(self.table_key_ref, TableKeyParameter)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        if self.table_key_snref is not None:
            self._table_key = resolve_snref(self.table_key_snref, odxrequire(context.parameters),
                                            TableKeyParameter)

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:

        if not isinstance(physical_value, (tuple, list)) or \
           len(physical_value) != 2 or \
           not isinstance(physical_value[0], str):
            odxraise(
                f"The physical value of TableStructParameter 'self.short_name' "
                f"must be a tuple containing the short name of the selected table "
                f"row as the first element and the physical value for the "
                f"row's structure or DOP as the second.", EncodeError)

        tr_short_name = physical_value[0]

        # make sure that the same table row is selected for all
        # TABLE-STRUCT parameters that are using the same key
        tk_short_name = self.table_key.short_name
        tk_value = encode_state.table_keys.get(tk_short_name)
        if tk_value is None:
            # no value for the key has been set yet. Set it to the
            # value which we are using right now
            encode_state.table_keys[tk_short_name] = tr_short_name
        elif tk_value != tr_short_name:
            odxraise(
                f"Cannot determine a unique value for table key '{tk_short_name}':  "
                f"Requested are '{tk_value}' and '{tr_short_name}'", EncodeError)
            return

        # deal with the static case (i.e., the table row is selected
        # by the table key object itself)
        if self.table_key.table_row is not None:
            if tr_short_name is not None and self.table_key.table_row.short_name != tr_short_name:
                odxraise(
                    f"The selected table row for the {self.short_name} "
                    f"parameter must be '{self.table_key.table_row.short_name}' "
                    f"instead of '{tr_short_name}'", EncodeError)
                return

            tr_short_name = self.table_key.table_row.short_name

        # encode the user specified value using the structure (or DOP)
        # of the selected table row
        table = self.table_key.table
        candidate_trs = [tr for tr in table.table_rows if tr.short_name == tr_short_name]
        if len(candidate_trs) == 0:
            odxraise(
                f"Could not find a table row named "
                f"'{tr_short_name}' in table '{table.short_name}'", EncodeError)
            return
        elif len(candidate_trs) > 1:
            odxraise(
                f"Found multiple table rows named "
                f"'{tr_short_name}' in table '{table.short_name}'", EncodeError)

        tr = candidate_trs[0]
        tr_value = physical_value[1]

        if tr.structure is not None:
            # the selected table row references a structure
            tr.structure.encode_into_pdu(tr_value, encode_state)
            return

        # if the table row does not reference a structure, it must
        # point to a DOP!
        if tr.dop is None:
            odxraise(f"Neither a structure nor a DOP has been defined for table row"
                     f"'{tr.short_name}'")
            return

        tr.dop.encode_into_pdu(tr_value, encode_state)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        # find the selected table row
        key_name = self.table_key.short_name

        decode_state.table_keys[key_name]
        table_row = decode_state.table_keys.get(key_name)
        if table_row is None:
            odxraise(
                f"No table key '{key_name}' found when decoding "
                f"table struct parameter '{str(self.short_name)}'", DecodeError)
            dummy_val = cast(str, None), cast(int, None)
            return dummy_val

        # Use DOP or structure to decode the value
        if table_row.dop is not None:
            dop = table_row.dop
            val = dop.decode_from_pdu(decode_state)
            return (table_row.short_name, val)
        elif table_row.structure is not None:
            val = table_row.structure.decode_from_pdu(decode_state)
            return (table_row.short_name, val)
        else:
            # the table row associated with the key neither defines a
            # DOP nor a structure -> ignore it
            return (table_row.short_name, cast(int, None))



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

from .odxtypes import ParameterValue, ParameterValueDict

if TYPE_CHECKING:
    from .diagservice import DiagService
    from .request import Request
    from .response import Response


@dataclass(kw_only=True)
class Message:
    """A diagnostic message with its interpretation.

    The `coded_message` attribute contains the binary data that's send
    over the wire using ISO-TP (CAN/LIN) or DoIP (Ethernet), while the
    remaining attributes of the class specify the "human readable"
    interpretation of the same data.
    """

    coded_message: bytes
    service: "DiagService"
    coding_object: Union["Request", "Response"]
    param_dict: ParameterValueDict

    def __getitem__(self, key: str) -> ParameterValue:
        return self.param_dict[key]



# SPDX-License-Identifier: MIT
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import odxassert, odxraise, odxrequire
from .field import Field
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import ParameterValue
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class StaticField(Field):
    """Array of a fixed number of structure objects"""
    fixed_number_of_items: int
    item_byte_size: int

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "StaticField":
        kwargs = dataclass_fields_asdict(Field.from_et(et_element, context))

        fixed_number_of_items = int(odxrequire(et_element.findtext('FIXED-NUMBER-OF-ITEMS')))
        item_byte_size = int(odxrequire(et_element.findtext('ITEM-BYTE-SIZE')))

        return StaticField(
            fixed_number_of_items=fixed_number_of_items, item_byte_size=item_byte_size, **kwargs)

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()
        return odxlinks

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

    @override
    def encode_into_pdu(self, physical_value: ParameterValue, encode_state: EncodeState) -> None:

        if not isinstance(physical_value,
                          Sequence) or len(physical_value) != self.fixed_number_of_items:
            odxraise(f"Value for static field '{self.short_name}' "
                     f"must be a list of size {self.fixed_number_of_items}")

        orig_is_end_of_pdu = encode_state.is_end_of_pdu
        encode_state.is_end_of_pdu = False
        for i, val in enumerate(physical_value):
            if not isinstance(val, dict):
                odxraise(f"The individual parameter values for static field '{self.short_name}' "
                         f"must be dictionaries for structure '{self.structure.short_name}'")

            if i == len(physical_value) - 1:
                encode_state.is_end_of_pdu = orig_is_end_of_pdu

            pos_before = encode_state.cursor_byte_position
            self.structure.encode_into_pdu(val, encode_state)
            pos_after = encode_state.cursor_byte_position

            if pos_after - pos_before > self.item_byte_size:
                odxraise(
                    f"Insufficient item byte size for static field {self.short_name}: "
                    f"Is {self.item_byte_size} bytes, but need at least {pos_after - pos_before} bytes"
                )
                encode_state.cursor_byte_position = pos_before + self.item_byte_size
            elif pos_after - pos_before < self.item_byte_size:
                # add some padding bytes
                encode_state.emplace_bytes(b'\x00' * (self.item_byte_size -
                                                      (pos_after - pos_before)))

        encode_state.is_end_of_pdu = orig_is_end_of_pdu

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:

        odxassert(decode_state.cursor_bit_position == 0,
                  "No bit position can be specified for static length fields!")

        orig_origin = decode_state.origin_byte_position
        decode_state.origin_byte_position = decode_state.cursor_byte_position

        result: list[ParameterValue] = []
        for _ in range(self.fixed_number_of_items):
            orig_cursor = decode_state.cursor_byte_position

            if decode_state.cursor_byte_position - orig_cursor > self.item_byte_size:
                odxraise(f"Insufficient item byte size for static field {self.short_name}: "
                         f"Is {self.item_byte_size} bytes, but need at least "
                         f"{decode_state.cursor_byte_position - orig_cursor} bytes")

            result.append(self.structure.decode_from_pdu(decode_state))

            decode_state.cursor_byte_position = orig_cursor + self.item_byte_size

        decode_state.origin_byte_position = orig_origin

        return result



# SPDX-License-Identifier: MIT
from xml.etree import ElementTree

from .diagcodedtype import DiagCodedType
from .exceptions import odxraise
from .globals import xsi
from .leadinglengthinfotype import LeadingLengthInfoType
from .minmaxlengthtype import MinMaxLengthType
from .odxdoccontext import OdxDocContext
from .paramlengthinfotype import ParamLengthInfoType
from .standardlengthtype import StandardLengthType


def create_any_diag_coded_type_from_et(et_element: ElementTree.Element,
                                       context: OdxDocContext) -> DiagCodedType:
    dct_type = et_element.get(f"{xsi}type")
    if dct_type == "LEADING-LENGTH-INFO-TYPE":
        return LeadingLengthInfoType.from_et(et_element, context)
    elif dct_type == "MIN-MAX-LENGTH-TYPE":
        return MinMaxLengthType.from_et(et_element, context)
    elif dct_type == "PARAM-LENGTH-INFO-TYPE":
        return ParamLengthInfoType.from_et(et_element, context)
    elif dct_type == "STANDARD-LENGTH-TYPE":
        return StandardLengthType.from_et(et_element, context)

    odxraise(f"Unknown DIAG-CODED-TYPE {dct_type}", NotImplementedError)
    return DiagCodedType.from_et(et_element, context)



# SPDX-License-Identifier: MIT
from collections.abc import Callable, Iterable
from enum import Enum
from typing import TYPE_CHECKING, Any, SupportsBytes, Union, overload
from xml.etree import ElementTree

from .exceptions import odxassert, odxraise, odxrequire

if TYPE_CHECKING:
    from odxtools.diagnostictroublecode import DiagnosticTroubleCode
    from odxtools.parameters.parameter import Parameter


def bytefield_to_bytearray(bytefield: str) -> bytearray:
    bytes_string = [bytefield[i:i + 2] for i in range(0, len(bytefield), 2)]
    return bytearray([int(x, 16) for x in bytes_string])


BytesTypes = bytearray | bytes | SupportsBytes
AtomicOdxType = str | int | float | BytesTypes

# dictionary mapping short names to a Parameter that needs to be
# specified. Complex parameters (structures) may contain
# sub-parameters, so this is a recursive type...
ParameterDict = dict[str, Union["Parameter", "ParameterDict"]]

# Dictionary mapping short names of parameters to the value it
# exhibits. Complex parameters (structures) may contain
# sub-parameters, so this is a recursive type, and fields encompass
# multiple items, so this can be a list of objects.
TableStructParameterValue = tuple[str, "ParameterValue"]
ParameterValue = Union[AtomicOdxType, "ParameterValueDict", TableStructParameterValue,
                       Iterable["ParameterValue"], "DiagnosticTroubleCode"]
ParameterValueDict = dict[str, ParameterValue]


@overload
def odxstr_to_bool(str_val: None) -> None:
    ...


@overload
def odxstr_to_bool(str_val: str) -> bool:
    ...


def odxstr_to_bool(str_val: str | None) -> bool | None:
    if str_val is None:
        return None

    str_val = str_val.strip()
    odxassert(str_val in [
        "0",
        "1",
        "false",
        "true",
    ], f"String '{str_val}' cannot be converted to a boolean")

    return str_val in ["1", "true"]


def bool_to_odxstr(bool_val: bool) -> str:
    return "true" if bool_val else "false"


def parse_int(value: str) -> int:
    try:
        return int(value, 0)
    except (ValueError, TypeError):
        try:
            v = float(value)
        except Exception as e:
            odxraise(f"Error parsing numerical value '{value}': {e}")

        if not v.is_integer():
            odxraise(f"Expected an integer value, got {v}")
        return int(v)


#: conversion functions for strings from the XML to the types stored
#: by the internalized database
_PARSE_ODX_TYPE: dict[str, Callable[[str], AtomicOdxType]] = {
    "A_INT32": parse_int,
    "A_UINT32": parse_int,
    "A_FLOAT32": float,
    "A_FLOAT64": float,
    "A_UNICODE2STRING": str,
    "A_BYTEFIELD": bytefield_to_bytearray,
    "A_ASCIISTRING": str,
    "A_UTF8STRING": str,
}

#: mapping from type name strings specified by the XML to the types
#: used by the internalized database
_ODX_TYPE_TO_PYTHON_TYPE: dict[str, type[int | float | str | bytearray]] = {
    "A_INT32": int,
    "A_UINT32": int,
    "A_FLOAT32": float,
    "A_FLOAT64": float,
    "A_UNICODE2STRING": str,
    "A_BYTEFIELD": bytearray,
    "A_ASCIISTRING": str,
    "A_UTF8STRING": str,
}


def compare_odx_values(a: AtomicOdxType, b: AtomicOdxType) -> int:
    # this function implements the comparison according to the ODX
    # specification. (cf section 7.3.6.5)

    # numeric values are compared numerically (duh!)
    if isinstance(a, (int, float)):
        if not isinstance(b, (int, float)):
            odxraise()

        tmp = a - b
        if tmp < 0:
            return -1
        elif tmp > 0:
            return 1
        return 0

    # strings are compared lexicographically. (the spec only allows
    # equals, but this cannot easily implemented using a single
    # comparison function.
    if isinstance(a, str):
        if not isinstance(b, str):
            odxraise()

        if a < b:
            return -1
        elif b < a:
            return 1
        else:
            return 0

    # bytefields are treated like long integers: to pad the shorter
    # object with zeros and treat the results like strings.
    if isinstance(a, BytesTypes):
        if not isinstance(b, BytesTypes):
            odxraise()

        a_bytes = bytes(a)
        b_bytes = bytes(b)

        obj_len = max(len(a_bytes), len(b_bytes))

        tmp_a = a_bytes.ljust(obj_len, b'\x00')
        tmp_b = b_bytes.ljust(obj_len, b'\x00')

        if tmp_a > tmp_b:
            return 1
        elif tmp_a < tmp_b:
            return -1
        else:
            return 0

    odxraise(f"Unhandled comparsion between objects of type {type(a).__name__} "
             f"and {type(b).__name__}")


# format specifiers for the data type using the bitstruct module
_BITSTRUCT_FORMAT_LETTER_MAP__ = {
    "A_INT32": "u",  # the sign must be handled separately
    "A_UINT32": "u",
    "A_FLOAT32": "f",
    "A_FLOAT64": "f",
    "A_BYTEFIELD": "r",
    "A_UNICODE2STRING": "r",  # UTF-16 strings must be converted explicitly
    "A_ASCIISTRING": "r",
    "A_UTF8STRING": "r",
}


class DataType(Enum):
    """Types for the physical and internal value.

    These types can be used either as BASE-DATA-TYPE (for the "internal value")
    or the PHYSICAL-DATA-TYPE (for the "physical value").

    Relevant pages in the ASAM MCD-2D specification:
    * p. 38 (Table 1): ASAM types correspondence with XML Schema types
    * p. 96: Restrictions to the bit length given the BASE-DATA-TYPE.
    """

    A_INT32 = "A_INT32"
    A_UINT32 = "A_UINT32"
    A_FLOAT32 = "A_FLOAT32"
    A_FLOAT64 = "A_FLOAT64"
    A_UNICODE2STRING = "A_UNICODE2STRING"
    A_BYTEFIELD = "A_BYTEFIELD"

    # these two enums are only used by internal data type objects (DATA-TYPE)
    # not by the ones for physical values (PHYSICAL-DATA-TYPE)
    A_ASCIISTRING = "A_ASCIISTRING"
    A_UTF8STRING = "A_UTF8STRING"

    @property
    def python_type(self) -> type[int | float | str | bytearray]:
        return _ODX_TYPE_TO_PYTHON_TYPE[self.value]

    @property
    def bitstruct_format_letter(self) -> str:
        return _BITSTRUCT_FORMAT_LETTER_MAP__[self.value]

    def from_string(self, value: str) -> AtomicOdxType:
        return _PARSE_ODX_TYPE[self.value](value)

    @overload
    def create_from_et(self, et_element: None) -> None:
        ...

    @overload
    def create_from_et(self, et_element: ElementTree.Element) -> AtomicOdxType:
        ...

    def create_from_et(self, et_element: ElementTree.Element | None) -> AtomicOdxType | None:
        """
            Parse a V/VT value union and return an AtomicOdxType from them that match current datatype
            this includes, but not limited to COMPU-CONST, COMPU-DEFAULT-VALUE, COMPU-INVERSE-VALUE
        """
        if et_element is None:
            return None
        if (vt_elem := et_element.find("VT")) is not None:
            return self.from_string(odxrequire(vt_elem.text))
        elif (v_elem := et_element.find("V")) is not None:
            return self.from_string(odxrequire(v_elem.text))
        odxraise('Either V or VT needs to be specified')

    def make_from(self, value: Any) -> AtomicOdxType:
        if isinstance(value, str):
            # parse the string
            return self.from_string(value)
        else:
            # regular type cast of python objects
            return self.python_type(value)

    def isinstance(self, value: Any) -> bool:
        expected_type = self.python_type
        if isinstance(value, expected_type):
            return True
        elif expected_type is float and isinstance(value, (int, float)):
            return True
        elif self == DataType.A_BYTEFIELD and isinstance(value, BytesTypes):
            return True
        else:
            return False

    def __str__(self) -> str:
        if self == DataType.A_INT32:
            return "int"
        elif self == DataType.A_UINT32:
            return "uint"
        elif self in (DataType.A_FLOAT32, DataType.A_FLOAT64):
            return "float"
        elif self == DataType.A_BYTEFIELD:
            return "bytefield"
        elif self in (DataType.A_UNICODE2STRING, DataType.A_ASCIISTRING, DataType.A_UTF8STRING):
            return "string"
        else:
            return f"<unknown type '{self.value}'>"



# SPDX-License-Identifier: MIT
from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast
from xml.etree import ElementTree

from ..diagvariable import DiagVariable
from ..exceptions import odxassert
from ..nameditemlist import NamedItemList
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkRef
from ..snrefcontext import SnRefContext
from ..variablegroup import VariableGroup
from .diaglayer import DiagLayer
from .ecushareddataraw import EcuSharedDataRaw

if TYPE_CHECKING:
    from ..database import Database


@dataclass(kw_only=True)
class EcuSharedData(DiagLayer):
    """This is a diagnostic layer for data shared across others
    """

    @property
    def ecu_shared_data_raw(self) -> EcuSharedDataRaw:
        return cast(EcuSharedDataRaw, self.diag_layer_raw)

    @property
    def diag_variables_raw(self) -> list[OdxLinkRef | DiagVariable]:
        return self.ecu_shared_data_raw.diag_variables_raw

    @property
    def diag_variables(self) -> NamedItemList[DiagVariable]:
        return self.ecu_shared_data_raw.diag_variables

    @property
    def variable_groups(self) -> NamedItemList[VariableGroup]:
        return self.ecu_shared_data_raw.variable_groups

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "EcuSharedData":
        ecu_shared_data_raw = EcuSharedDataRaw.from_et(et_element, context)

        return EcuSharedData(diag_layer_raw=ecu_shared_data_raw)

    def __post_init__(self) -> None:
        super().__post_init__()

        odxassert(
            isinstance(self.diag_layer_raw, EcuSharedDataRaw),
            "The raw diagnostic layer passed to EcuSharedData "
            "must be a EcuSharedDataRaw")

    def _finalize_init(self, database: "Database", odxlinks: OdxLinkDatabase) -> None:
        """This method makes sure that all references in sub-objects
        of the ECU shared data layer are resolved

        In contrast to hierarchy element layers, ECU shared data
        layers do not need to deal with inheritance...

        """

        # this attribute may be removed later. it is currently
        # required to properly deal with auxiliary files within the
        # diagnostic layer.
        self._database = database

        #####
        # resolve all SNREFs. TODO: We allow SNREFS to objects that
        # were inherited by the diaglayer. This might not be allowed
        # by the spec (So far, I haven't found any definitive
        # statement...)
        #####
        context = SnRefContext(database=database)
        context.diag_layer = self
        self._resolve_snrefs(context)
        context.diag_layer = None

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        """Create a deep copy of the protocol layer

        Note that the copied diagnostic layer is not fully
        initialized, so `_finalize_init()` should to be called on it
        before it can be used normally.
        """

        result = super().__deepcopy__(memo)

        # note that the self.ecu_shared_data_raw object is *not* copied at
        # this place because the attribute points to the same object
        # as self.diag_layer_raw.
        result.ecu_shared_data_raw = deepcopy(self.ecu_shared_data_raw, memo)

        return result



# SPDX-License-Identifier: MIT
import re
import warnings
from collections.abc import Callable, Iterable
from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast
from xml.etree import ElementTree

from ..additionalaudience import AdditionalAudience
from ..admindata import AdminData
from ..comparaminstance import ComparamInstance
from ..diagcomm import DiagComm
from ..diagdatadictionaryspec import DiagDataDictionarySpec
from ..diagservice import DiagService
from ..exceptions import OdxWarning, odxassert, odxraise
from ..functionalclass import FunctionalClass
from ..nameditemlist import NamedItemList, OdxNamed
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId
from ..parentref import ParentRef
from ..response import Response
from ..singleecujob import SingleEcuJob
from ..snrefcontext import SnRefContext
from ..specialdatagroup import SpecialDataGroup
from ..statechart import StateChart
from ..unitgroup import UnitGroup
from ..unitspec import UnitSpec
from .diaglayer import DiagLayer
from .hierarchyelementraw import HierarchyElementRaw

if TYPE_CHECKING:
    from .database import Database
    from .protocol import Protocol

TNamed = TypeVar("TNamed", bound=OdxNamed)


@dataclass(kw_only=True)
class HierarchyElement(DiagLayer):
    """This is the base class for diagnostic layers that may be involved in value inheritance
    """

    @property
    def hierarchy_element_raw(self) -> HierarchyElementRaw:
        return cast(HierarchyElementRaw, self.diag_layer_raw)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "HierarchyElement":
        hierarchy_element_raw = HierarchyElementRaw.from_et(et_element, context)

        return HierarchyElement(diag_layer_raw=hierarchy_element_raw)

    def __post_init__(self) -> None:
        super().__post_init__()

        self._global_negative_responses: NamedItemList[Response]

        odxassert(
            isinstance(self.diag_layer_raw, HierarchyElementRaw),
            "The raw diagnostic layer passed to HierarchyElement "
            "must be a HierarchyElementRaw")

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        """Create a deep copy of the hierarchy element

        Note that the copied diagnostic layer is not fully
        initialized, so `_finalize_init()` should to be called on it
        before it can be used normally.
        """

        new_he = super().__deepcopy__(memo)

        # note that the self.hierarchy_element_raw object is *not*
        # copied at this place because the attribute points to the
        # same object as self.diag_layer_raw.
        new_he.hierarchy_element_raw = deepcopy(self.hierarchy_element_raw)

        return new_he

    def _finalize_init(self, database: "Database", odxlinks: OdxLinkDatabase) -> None:
        """This method deals with everything inheritance related and
        -- after the final set of objects covered by the diagnostic
        layer is determined -- resolves any short name references in
        the diagnostic layer.

        TODO: In some corner cases, the short name resolution is not
        correct: E.g. Given three layers A, B, and C, where B and C
        derive from A and A defines the diagnostic communication
        SA. If now B overrides SA and there are short name references
        to SA in A, the object to which this reference is resolved is
        undefined. An easy fix for this problem is to copy all
        inherited objects in derived layers, but that would lead to
        excessive memory consumption for large databases...
        """

        #####
        # fill in all applicable objects that use value inheritance
        #####

        self._compute_value_inheritance(odxlinks)

        ############
        # create a new unit_spec object. This is necessary because
        # unit_groups are subject to value inheritance.

        # unit groups applicable to this diaglayer (i.e., including
        # value inheritance)
        unit_groups = self._compute_available_unit_groups()

        # convenience variable for the locally-defined unit spec
        local_unit_spec: UnitSpec | None = None
        if self.diag_layer_raw.diag_data_dictionary_spec is not None:
            local_unit_spec = self.diag_layer_raw.diag_data_dictionary_spec.unit_spec
        else:
            local_unit_spec = None

        unit_spec: UnitSpec | None = None
        if local_unit_spec is None and not unit_groups:
            # no locally defined unit spec and no inherited unit groups
            unit_spec = None
        elif local_unit_spec is None:
            # no locally defined unit spec but inherited unit groups
            unit_spec = UnitSpec(
                unit_groups=NamedItemList(unit_groups),
                units=NamedItemList([]),
                physical_dimensions=NamedItemList([]),
                admin_data=None,
                sdgs=[])
        else:
            # locally defined unit spec and inherited unit groups
            unit_spec = UnitSpec(
                unit_groups=NamedItemList(unit_groups),
                units=local_unit_spec.units,
                physical_dimensions=local_unit_spec.physical_dimensions,
                admin_data=None,
                sdgs=[])
        ############

        dops = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.data_object_props,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        structures = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.structures,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        dtc_dops = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.dtc_dops,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        static_fields = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.static_fields,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        end_of_pdu_fields = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.end_of_pdu_fields,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        dynamic_endmarker_fields = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.dynamic_endmarker_fields,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        dynamic_length_fields = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.dynamic_length_fields,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        env_data_descs = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.env_data_descs,
            lambda parent_ref: parent_ref.not_inherited_dops,
        )
        env_datas = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.env_datas, lambda parent_ref: parent_ref.not_inherited_dops)
        muxs = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.muxs, lambda parent_ref: parent_ref.not_inherited_dops)
        tables = self._compute_available_ddd_spec_items(
            lambda ddd_spec: ddd_spec.tables, lambda parent_ref: parent_ref.not_inherited_tables)

        ddds_admin_data: AdminData | None = None
        ddds_sdgs: list[SpecialDataGroup] = []
        if self.diag_layer_raw.diag_data_dictionary_spec:
            ddds_admin_data = self.diag_layer_raw.diag_data_dictionary_spec.admin_data
            ddds_sdgs = self.diag_layer_raw.diag_data_dictionary_spec.sdgs

        # create a DiagDataDictionarySpec which includes all the
        # inherited objects. To me, this seems rather inelegant, but
        # hey, it's described like this in the standard.
        self._diag_data_dictionary_spec = DiagDataDictionarySpec(
            admin_data=ddds_admin_data,
            data_object_props=dops,
            dtc_dops=dtc_dops,
            structures=structures,
            static_fields=static_fields,
            end_of_pdu_fields=end_of_pdu_fields,
            dynamic_endmarker_fields=dynamic_endmarker_fields,
            dynamic_length_fields=dynamic_length_fields,
            tables=tables,
            env_data_descs=env_data_descs,
            env_datas=env_datas,
            muxs=muxs,
            unit_spec=unit_spec,
            sdgs=ddds_sdgs,
        )

        #####
        # compute the communication parameters applicable to the
        # diagnostic layer. Note that communication parameters do
        # *not* use value inheritance, but a slightly different
        # scheme, cf the docstring of
        # _compute_available_commmunication_parameters().
        #####
        self._comparam_refs = NamedItemList(self._compute_available_commmunication_parameters())

        #####
        # resolve all SNREFs. TODO: We allow SNREFS to objects that
        # were inherited by the diaglayer. This might not be allowed
        # by the spec (So far, I haven't found any definitive
        # statement...)
        #####
        context = SnRefContext(database=database)
        context.diag_layer = self
        self._resolve_snrefs(context)
        context.diag_layer = None

    #####
    # <value inheritance mechanism helpers>
    #####
    def _compute_value_inheritance(self, odxlinks: OdxLinkDatabase) -> None:
        # diagnostic communication objects with the ODXLINKs resolved
        diag_comms = self._compute_available_diag_comms(odxlinks)
        self._diag_comms = NamedItemList[DiagComm](diag_comms)

        # filter the diag comms for services and single-ECU jobs
        diag_services = [dc for dc in diag_comms if isinstance(dc, DiagService)]
        single_ecu_jobs = [dc for dc in diag_comms if isinstance(dc, SingleEcuJob)]
        self._diag_services = NamedItemList(diag_services)
        self._single_ecu_jobs = NamedItemList(single_ecu_jobs)

        global_negative_responses = self._compute_available_global_neg_responses(odxlinks)
        self._global_negative_responses = NamedItemList(global_negative_responses)

        functional_classes = self._compute_available_functional_classes()
        self._functional_classes = NamedItemList(functional_classes)

        additional_audiences = self._compute_available_additional_audiences()
        self._additional_audiences = NamedItemList(additional_audiences)

        state_charts = self._compute_available_state_charts()
        self._state_charts = NamedItemList(state_charts)

    def _get_parent_refs_sorted_by_priority(self, reverse: bool = False) -> Iterable[ParentRef]:
        return sorted(
            getattr(self.diag_layer_raw, "parent_refs", []),
            key=lambda pr: pr.layer.variant_type.inheritance_priority,
            reverse=reverse)

    def _compute_available_objects(
        self,
        get_local_objects: Callable[["DiagLayer"], Iterable[TNamed]],
        get_not_inherited: Callable[[ParentRef], Iterable[str]],
    ) -> Iterable[TNamed]:
        """Helper method to compute the set of all objects applicable
        to the DiagLayer if these objects are subject to the value
        inheritance mechanism

        Note that all objects subject to the value inheritance
        mechanism exhibit a short_name attribute.

        :param get_local_objects: Function mapping a DiagLayer to the
        set of objects that are locally defined by that DiagLayer. If
        any of these objects have already been defined in any of the
        parents, the locally defined instance overrides them.

        :param get_not_inherited: Function mapping a ParentRef to the
        set of short names of the objects which shall not be inherited
        from the parents.

        """

        local_objects = get_local_objects(self)
        local_object_short_names = {x.short_name for x in local_objects}
        result_dict: dict[str, tuple[TNamed, DiagLayer]] = {}

        # populate the result dictionary with the inherited objects
        for parent_ref in self._get_parent_refs_sorted_by_priority(reverse=True):
            parent_dl = parent_ref.layer

            # retrieve the set of short names of the objects which we
            # are not supposed to inherit
            not_inherited_short_names = set(get_not_inherited(parent_ref))

            # compute the list of objects which we are supposed to
            # inherit from this diagnostic layer
            inherited_objects = [
                x
                for x in parent_dl._compute_available_objects(get_local_objects, get_not_inherited)
                if x.short_name not in not_inherited_short_names
            ]

            # update the result set with the objects from the current parent_ref
            for obj in inherited_objects:

                # no object with the given short name currently
                # exits. add it to the result set and continue
                if obj.short_name not in result_dict:
                    result_dict[obj.short_name] = (obj, parent_dl)
                    continue

                # if an object with a given name already exists,
                # there's no problem if it was inherited from a parent
                # of different priority than the one currently
                # considered
                orig_prio = result_dict[obj.short_name][1].variant_type.inheritance_priority
                new_prio = parent_dl.variant_type.inheritance_priority
                if new_prio < orig_prio:
                    continue
                elif orig_prio < new_prio:
                    result_dict[obj.short_name] = (obj, parent_dl)
                    continue

                # if there is a conflict on the same priority level,
                # it does not matter if the object is overridden
                # locally anyway...
                if obj.short_name in local_object_short_names:
                    continue

                # if all of these conditions do not apply, and if the
                # inherited objects are identical, there is no
                # conflict. (note that value comparisons of complete
                # complex objects tend to be expensive, so this test
                # is done last.)
                if obj == result_dict[obj.short_name][0]:
                    continue

                odxraise(f"Diagnostic layer {self.short_name} cannot inherit object "
                         f"{obj.short_name} due to an unresolveable inheritance conflict between "
                         f"parent layers {result_dict[obj.short_name][1].short_name} "
                         f"and {parent_dl.short_name}")

        # add the locally defined entries, overriding the inherited
        # ones if necessary
        for obj in local_objects:
            result_dict[obj.short_name] = (obj, self)

        return [x[0] for x in result_dict.values()]

    def _compute_available_diag_comms(self, odxlinks: OdxLinkDatabase) -> Iterable[DiagComm]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[DiagComm]:
            return dl._get_local_diag_comms(odxlinks)

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return parent_ref.not_inherited_diag_comms

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_global_neg_responses(self, odxlinks: OdxLinkDatabase) \
            -> Iterable[Response]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[Response]:
            return dl.diag_layer_raw.global_negative_responses

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return parent_ref.not_inherited_global_neg_responses

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_ddd_spec_items(
        self,
        include: Callable[[DiagDataDictionarySpec], Iterable[TNamed]],
        exclude: Callable[["ParentRef"], list[str]],
    ) -> NamedItemList[TNamed]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[TNamed]:
            if dl.diag_layer_raw.diag_data_dictionary_spec is None:
                return []
            return include(dl.diag_layer_raw.diag_data_dictionary_spec)

        found = self._compute_available_objects(get_local_objects_fn, exclude)
        return NamedItemList(found)

    def _compute_available_functional_classes(self) -> Iterable[FunctionalClass]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[FunctionalClass]:
            return dl.diag_layer_raw.functional_classes

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return []

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_additional_audiences(self) -> Iterable[AdditionalAudience]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[AdditionalAudience]:
            return dl.diag_layer_raw.additional_audiences

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return []

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_state_charts(self) -> Iterable[StateChart]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[StateChart]:
            return dl.diag_layer_raw.state_charts

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return []

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_unit_groups(self) -> Iterable[UnitGroup]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[UnitGroup]:
            return dl._get_local_unit_groups()

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return []

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    #####
    # </value inheritance mechanism helpers>
    #####

    #######
    # <properties subject to value inheritance>
    #######
    @property
    def diag_data_dictionary_spec(self) -> DiagDataDictionarySpec:
        return self._diag_data_dictionary_spec

    @property
    def diag_comms(self) -> NamedItemList[DiagComm]:
        """All diagnostic communication primitives applicable to this DiagLayer

        Diagnostic communication primitives are diagnostic services as
        well as single-ECU jobs. This list has all references
        resolved.
        """
        return self._diag_comms

    @property
    def services(self) -> NamedItemList[DiagService]:
        """This property is an alias for `.diag_services`"""
        return self._diag_services

    @property
    def diag_services(self) -> NamedItemList[DiagService]:
        """All diagnostic services applicable to this DiagLayer

        This is a subset of all diagnostic communication
        primitives. All references are resolved in the list returned.
        """
        return self._diag_services

    @property
    def single_ecu_jobs(self) -> NamedItemList[SingleEcuJob]:
        """All single-ECU jobs applicable to this DiagLayer

        This is a subset of all diagnostic communication
        primitives. All references are resolved in the list returned.
        """
        return self._single_ecu_jobs

    @property
    def global_negative_responses(self) -> NamedItemList[Response]:
        """All global negative responses applicable to this DiagLayer"""
        return self._global_negative_responses

    @property
    def functional_classes(self) -> NamedItemList[FunctionalClass]:
        """All functional classes applicable to this DiagLayer"""
        return self._functional_classes

    @property
    def state_charts(self) -> NamedItemList[StateChart]:
        """All state charts applicable to this DiagLayer"""
        return self._state_charts

    @property
    def additional_audiences(self) -> NamedItemList[AdditionalAudience]:
        """All audiences applicable to this DiagLayer"""
        return self._additional_audiences

    #######
    # </properties subject to value inheritance>
    #######

    #####
    # <communication parameter handling>
    #####
    def _compute_available_commmunication_parameters(self) -> list[ComparamInstance]:
        """Compute the list of communication parameters that apply to
        the diagnostic layer

        Be aware that the inheritance scheme for communication
        parameters is slightly different than for objects that are
        subject to value inheritance:

        - The ODXLINK ID id of communication parameters is used to
          override inherited parameters instead of the short name.
        - A parameter is only overridden if the specified protocol
          matches.

        Note that the specification leaves some room for
        interpretation here: It says that if no protocol is specified,
        the parameter shall apply to any protocol. But what happens if
        the the same comparam is specified with and without a
        protocol? Is this allowed at all? If yes, which of these
        definitions gets priority? How does this interact with
        inheritance? The approach taken here is to allow such cases
        and to use the specific comparams if possible whilst the ones
        without a specified protocol are taken as fallbacks...

        """
        com_params_dict: dict[tuple[str, str | None], ComparamInstance] = {}

        # Look in parent refs for inherited communication
        # parameters. First fetch the communication parameters from
        # low priority parents, then update with increasing priority.
        for parent_ref in self._get_parent_refs_sorted_by_priority():
            parent_layer = parent_ref.layer
            if not isinstance(parent_layer, HierarchyElement):
                continue
            for cp in parent_layer._compute_available_commmunication_parameters():
                com_params_dict[(cp.spec_ref.ref_id, cp.protocol_snref)] = cp

        # finally, handle the locally defined communication parameters
        for cp in getattr(self.hierarchy_element_raw, "comparam_refs", []):
            com_params_dict[(cp.spec_ref.ref_id, cp.protocol_snref)] = cp

        return list(com_params_dict.values())

    @property
    def comparam_refs(self) -> NamedItemList[ComparamInstance]:
        """All communication parameters applicable to this DiagLayer

        Note that, although communication parameters use inheritance,
        it is *not* the "value inheritance" scheme used by e.g. DOPs,
        tables, state charts, ...
        """
        return self._comparam_refs

    @cached_property
    def protocols(self) -> NamedItemList["Protocol"]:
        """Return the set of all protocols which are applicable to the diagnostic layer

        Note that protocols are *not* explicitly defined by the XML,
        but they are the parent layers of variant type "PROTOCOL".

        """
        from .protocol import Protocol

        result_dict: dict[str, Protocol] = {}

        for parent_ref in getattr(self, "parent_refs", []):
            for prot in getattr(parent_ref.layer, "protocols", []):
                result_dict[prot.short_name] = prot

        if isinstance(self, Protocol):
            result_dict[self.diag_layer_raw.short_name] = self

        return NamedItemList(result_dict.values())

    def get_comparam(
        self,
        cp_short_name: str,
        *,
        protocol: Union[str, "Protocol"] | None = None,
    ) -> ComparamInstance | None:
        """Find a specific communication parameter according to some criteria.

        Setting a given parameter to `None` means "don't care"."""

        from .protocol import Protocol

        protocol_name: str | None = None
        if isinstance(protocol, Protocol):
            protocol_name = protocol.short_name
        else:
            protocol_name = protocol

        # determine the set of applicable communication parameters
        cps = [cp for cp in self.comparam_refs if cp.short_name == cp_short_name]
        if protocol_name is not None:
            cps = [cp for cp in cps if cp.protocol_snref in (None, protocol_name)]

        if len(cps) > 1:
            warnings.warn(
                f"Communication parameter `{cp_short_name}` specified more "
                f"than once. Using first occurence.",
                OdxWarning,
                stacklevel=1,
            )
        elif len(cps) == 0:
            return None

        return cps[0]

    def get_max_can_payload_size(self,
                                 protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """Return the maximum size of a CAN frame payload that can be
        transmitted in bytes.

        For classic CAN busses, this is basically always 8. CAN-FD can
        send up to 64 bytes per frame.

        """
        com_param = self.get_comparam("CP_CANFDTxMaxDataLength", protocol=protocol)
        if com_param is None:
            rx_id = self.get_can_receive_id(protocol=protocol)
            if rx_id is not None:
                # bus is CAN. We also use 8 bytes for CAN-FD if this
                # parameter was not specified.
                return 8

            # bus is not CAN
            return None

        val = com_param.value
        if not isinstance(val, str):
            return None

        m = re.search("TX_DL *= *([0-9]*)", val)
        if m:
            return int(m.group(1))

        # unexpected format of parameter value
        return 8

    def uses_can(self, protocol: Union[str, "Protocol"] | None = None) -> bool:
        """
        Check if CAN ought to be used as the link layer protocol.
        """
        return self.get_can_receive_id(protocol=protocol) is not None

    def uses_can_fd(self, protocol: Union[str, "Protocol"] | None = None) -> bool:
        """Check if CAN-FD ought to be used.

        If the ECU is not using CAN-FD for the specified protocol, `False`
        is returned.

        If this method returns `True`, `.uses_can() == True` is implied
        for the protocol.
        """

        if not self.uses_can(protocol):
            return False

        com_param = self.get_comparam("CP_CANFDTxMaxDataLength", protocol=protocol)
        if com_param is None:
            return False

        return "CANFD" in com_param.value

    def get_can_baudrate(self, protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """Baudrate of the CAN bus which is used by the ECU [bits/s]

        If the ECU is not using CAN for the specified protocol, None
        is returned.

        """
        com_param = self.get_comparam("CP_Baudrate", protocol=protocol)
        if com_param is None:
            return None

        val = com_param.value
        if not isinstance(val, str):
            return None

        return int(val)

    def get_can_fd_baudrate(self, protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """Data baudrate of the CAN bus which is used by the ECU [bits/s]

        If the ECU is not using CAN-FD for the specified protocol,
        None is returned.
        """
        if not self.uses_can_fd(protocol=protocol):
            return None

        com_param = self.get_comparam("CP_CANFDBaudrate", protocol=protocol)
        if com_param is None:
            return None

        val = com_param.value
        if not isinstance(val, str):
            return None

        return int(val)

    def get_can_receive_id(self, protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """CAN ID to which the ECU listens for diagnostic messages"""
        com_param = self.get_comparam("CP_UniqueRespIdTable", protocol=protocol)
        if com_param is None:
            return None

        with warnings.catch_warnings():
            # depending on the protocol, we may get
            # "Communication parameter 'CP_UniqueRespIdTable' does not
            # specify 'CP_CanPhysReqId'" warning here. we don't want this
            # warning and simply return None...
            warnings.simplefilter("ignore", category=OdxWarning)
            result = com_param.get_subvalue("CP_CanPhysReqId")
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_can_send_id(self, protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """CAN ID to which the ECU sends replies to diagnostic messages"""

        # this hopefully resolves to the 'CP_UniqueRespIdTable'
        # parameter from the ISO_15765_2 comparam subset. (There is a
        # parameter with the same name in the ISO_13400_2_DIS_2015
        # subset for DoIP. If the wrong one is retrieved, try
        # specifying the protocol.)
        com_param = self.get_comparam("CP_UniqueRespIdTable", protocol=protocol)
        if com_param is None:
            return None

        with warnings.catch_warnings():
            # depending on the protocol, we may get
            # "Communication parameter 'CP_UniqueRespIdTable' does not
            # specify 'CP_CanRespUSDTId'" warning here. we don't want this
            # warning and simply return None...
            warnings.simplefilter("ignore", category=OdxWarning)
            result = com_param.get_subvalue("CP_CanRespUSDTId")
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_can_func_req_id(self, protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """CAN Functional Request Id."""
        com_param = self.get_comparam("CP_CanFuncReqId", protocol=protocol)
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_doip_logical_ecu_address(self,
                                     protocol: Union[str, "Protocol"] | None = None) -> int | None:
        """Return the address of the ECU when using functional addressing.

        The parameter protocol is used to distinguish between
        different interfaces, e.g., offboard and onboard DoIP
        Ethernet.
        """

        # this hopefully resolves to the 'CP_UniqueRespIdTable'
        # parameter from the ISO_13400_2_DIS_2015 comparam
        # subset. (There is a parameter with the same name in the
        # ISO_15765_2 subset for CAN. If the wrong one is retrieved,
        # try specifying the protocol.)
        com_param = self.get_comparam("CP_UniqueRespIdTable", protocol=protocol)

        if com_param is None:
            return None

        # The CP_DoIPLogicalEcuAddress is specified by the
        # "CP_DoIPLogicalEcuAddress" subvalue of the complex comparam
        # CP_UniqueRespIdTable of the ISO_13400_2_DIS_2015 comparam
        # subset. Depending of the underlying transport protocol,
        # (i.e., CAN using ISO-TP) this subvalue might not exist.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=OdxWarning)
            ecu_addr = com_param.get_subvalue("CP_DoIPLogicalEcuAddress")
        if ecu_addr is None:
            return None
        return int(ecu_addr)

    def get_doip_logical_gateway_address(self,
                                         protocol: Union[str, "Protocol"] | None = None
                                        ) -> int | None:
        """The logical gateway address for the diagnosis over IP transport protocol"""

        # retrieve CP_DoIPLogicalGatewayAddress from the
        # ISO_13400_2_DIS_2015 subset. hopefully.
        com_param = self.get_comparam("CP_DoIPLogicalGatewayAddress", protocol=protocol)
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_doip_logical_tester_address(self,
                                        protocol: Union[str, "Protocol"] | None = None
                                       ) -> int | None:
        """DoIp logical gateway address"""

        # retrieve CP_DoIPLogicalTesterAddress from the
        # ISO_13400_2_DIS_2015 subset. hopefully.
        com_param = self.get_comparam("CP_DoIPLogicalTesterAddress", protocol=protocol)
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_doip_logical_functional_address(self,
                                            protocol: Union[str, "Protocol"] | None = None
                                           ) -> int | None:
        """The logical functional DoIP address of the ECU."""

        # retrieve CP_DoIPLogicalFunctionalAddress from the
        # ISO_13400_2_DIS_2015 subset. hopefully.
        com_param = self.get_comparam(
            "CP_DoIPLogicalFunctionalAddress",
            protocol=protocol,
        )
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_doip_routing_activation_timeout(self,
                                            protocol: Union[str, "Protocol"] | None = None
                                           ) -> float | None:
        """The timout for the DoIP routing activation request in seconds"""

        # retrieve CP_DoIPRoutingActivationTimeout from the
        # ISO_13400_2_DIS_2015 subset. hopefully.
        com_param = self.get_comparam("CP_DoIPRoutingActivationTimeout", protocol=protocol)
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return float(result) / 1e6

    def get_doip_routing_activation_type(self,
                                         protocol: Union[str, "Protocol"] | None = None
                                        ) -> int | None:
        """The DoIP routing activation type

        The number returned has the following meaning:

        0          Default
        1          WWH-OBD (worldwide harmonized onboard diagnostic).
        2-0xDF     reserved
        0xE0-0xFF  OEM-specific
        """

        # retrieve CP_DoIPRoutingActivationType from the
        # ISO_13400_2_DIS_2015 subset. hopefully.
        com_param = self.get_comparam("CP_DoIPRoutingActivationType", protocol=protocol)
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        return int(result)

    def get_tester_present_time(self,
                                protocol: Union[str, "Protocol"] | None = None) -> float | None:
        """Timeout on inactivity in seconds.

        This is defined by the communication parameter "CP_TesterPresentTime".

        Description of the comparam: "Time between a response and the
        next subsequent tester present message (if no other request is
        sent to this ECU) in case of physically addressed requests."
        """

        # retrieve CP_TesterPresentTime from either the
        # ISO_13400_2_DIS_2015 or the ISO_15765_2 subset.
        com_param = self.get_comparam("CP_TesterPresentTime", protocol=protocol)
        if com_param is None:
            return None

        result = com_param.get_value()
        if result is None:
            return None
        odxassert(isinstance(result, str))

        # the comparam specifies microseconds. convert this to seconds
        return float(result) / 1e6

    #####
    # </communication parameter handling>
    #####



# SPDX-License-Identifier: MIT
from xml.etree import ElementTree

from ..exceptions import odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import DataType
from .compucodecompumethod import CompuCodeCompuMethod
from .compumethod import CompuMethod
from .identicalcompumethod import IdenticalCompuMethod
from .linearcompumethod import LinearCompuMethod
from .ratfunccompumethod import RatFuncCompuMethod
from .scalelinearcompumethod import ScaleLinearCompuMethod
from .scaleratfunccompumethod import ScaleRatFuncCompuMethod
from .tabintpcompumethod import TabIntpCompuMethod
from .texttablecompumethod import TexttableCompuMethod


def create_any_compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                                    internal_type: DataType,
                                    physical_type: DataType) -> CompuMethod:
    compu_category = odxrequire(et_element.findtext("CATEGORY"))

    if compu_category == "IDENTICAL":
        return IdenticalCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "LINEAR":
        return LinearCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "SCALE-LINEAR":
        return ScaleLinearCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "RAT-FUNC":
        return RatFuncCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "SCALE-RAT-FUNC":
        return ScaleRatFuncCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "TEXTTABLE":
        return TexttableCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "COMPUCODE":
        return CompuCodeCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
    elif compu_category == "TAB-INTP":
        return TabIntpCompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)

    # TODO: Implement all categories (never instantiate the CompuMethod base class!)
    odxraise(f"Warning: Computation category {compu_category} is not implemented!")

    return IdenticalCompuMethod.compu_method_from_et(
        et_element, context, internal_type=internal_type, physical_type=physical_type)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .admindata import AdminData
from .audience import Audience
from .dataobjectproperty import DataObjectProperty
from .dtcdop import DtcDop
from .element import IdentifiableElement
from .exceptions import odxassert, odxraise, odxrequire
from .functionalclass import FunctionalClass
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .odxtypes import AtomicOdxType, odxstr_to_bool
from .preconditionstateref import PreConditionStateRef
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .statetransitionref import StateTransitionRef
from .structure import Structure
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .table import Table


@dataclass(kw_only=True)
class TableRow(IdentifiableElement):
    """This class represents a TABLE-ROW."""
    table_ref: OdxLinkRef
    key_raw: str

    # The spec mandates that either a structure or a non-complex DOP
    # must be referenced here, i.e., exactly one of the four
    # attributes below is not None
    dop_ref: OdxLinkRef | None = None
    dop_snref: str | None = None
    structure_ref: OdxLinkRef | None = None
    structure_snref: str | None = None

    sdgs: list[SpecialDataGroup] = field(default_factory=list)
    audience: Audience | None = None
    functional_class_refs: list[OdxLinkRef] = field(default_factory=list)
    state_transition_refs: list[StateTransitionRef] = field(default_factory=list)
    pre_condition_state_refs: list[PreConditionStateRef] = field(default_factory=list)
    admin_data: AdminData | None = None

    is_executable_raw: bool | None = None
    semantic: str | None = None
    is_mandatory_raw: bool | None = None
    is_final_raw: bool | None = None

    @property
    def table(self) -> "Table":
        return self._table

    # the value of the key expressed in the type represented by the
    # referenced DOP
    @property
    def key(self) -> AtomicOdxType | None:
        return self._key

    @property
    def dop(self) -> DataObjectProperty | None:
        """The data object property object resolved by dop_ref."""
        return self._dop

    @property
    def structure(self) -> Structure | None:
        """The structure associated with this table row."""
        return self._structure

    @property
    def functional_classes(self) -> NamedItemList[FunctionalClass]:
        return self._functional_classes

    @property
    def is_executable(self) -> bool:
        return self.is_executable_raw in (None, True)

    @property
    def is_mandatory(self) -> bool:
        return self.is_mandatory_raw is True

    @property
    def is_final(self) -> bool:
        return self.is_final_raw is True

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> Any:
        raise RuntimeError(
            "Calling TableRow.from_et() is not allowed. Use TableRow.tablerow_from_et().")

    @staticmethod
    def tablerow_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                         table_ref: OdxLinkRef) -> "TableRow":
        """Reads a TABLE-ROW."""
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        key_raw = odxrequire(et_element.findtext("KEY"))

        dop_ref = OdxLinkRef.from_et(et_element.find("DATA-OBJECT-PROP-REF"), context)
        dop_snref: str | None = None
        if (dop_snref_elem := et_element.find("DATA-OBJECT-PROP-SNREF")) is not None:
            dop_snref = dop_snref_elem.attrib["SHORT-NAME"]

        structure_ref = OdxLinkRef.from_et(et_element.find("STRUCTURE-REF"), context)
        structure_snref: str | None = None
        if (structure_snref_elem := et_element.find("STRUCTURE-SNREF")) is not None:
            structure_snref = structure_snref_elem.attrib["SHORT-NAME"]

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        audience = None
        if (audience_elem := et_element.find("AUDIENCE")) is not None:
            audience = Audience.from_et(audience_elem, context)

        functional_class_refs = [
            odxrequire(OdxLinkRef.from_et(el, context))
            for el in et_element.iterfind("FUNCT-CLASS-REFS/FUNCT-CLASS-REF")
        ]

        state_transition_refs = [
            StateTransitionRef.from_et(el, context)
            for el in et_element.iterfind("STATE-TRANSITION-REFS/STATE-TRANSITION-REF")
        ]

        pre_condition_state_refs = [
            PreConditionStateRef.from_et(el, context)
            for el in et_element.iterfind("PRE-CONDITION-STATE-REFS/PRE-CONDITION-STATE-REF")
        ]

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)

        is_executable_raw = odxstr_to_bool(et_element.attrib.get("IS-EXECUTABLE"))
        semantic = et_element.attrib.get("SEMANTIC")
        is_mandatory_raw = odxstr_to_bool(et_element.attrib.get("IS-MANDATORY"))
        is_final_raw = odxstr_to_bool(et_element.attrib.get("IS-FINAL"))

        return TableRow(
            table_ref=table_ref,
            key_raw=key_raw,
            dop_ref=dop_ref,
            dop_snref=dop_snref,
            structure_ref=structure_ref,
            structure_snref=structure_snref,
            sdgs=sdgs,
            audience=audience,
            functional_class_refs=functional_class_refs,
            state_transition_refs=state_transition_refs,
            pre_condition_state_refs=pre_condition_state_refs,
            admin_data=admin_data,
            is_executable_raw=is_executable_raw,
            semantic=semantic,
            is_mandatory_raw=is_mandatory_raw,
            is_final_raw=is_final_raw,
            **kwargs)

    def __post_init__(self) -> None:
        self._dop: DataObjectProperty | None = None
        self._structure: Structure | None = None

        n = sum([0 if x is None else 1 for x in (self.dop_ref, self.dop_snref)])
        odxassert(
            n <= 1,
            f"Table row {self.short_name}: The dop can either be defined using ODXLINK or SNREF but not both."
        )

        n = sum([0 if x is None else 1 for x in (self.structure_ref, self.structure_snref)])
        odxassert(
            n <= 1,
            f"Table row {self.short_name}: The structure can either be defined using ODXLINK or SNREF but not both."
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        if self.audience is not None:
            result.update(self.audience._build_odxlinks())

        for st_ref in self.state_transition_refs:
            result.update(st_ref._build_odxlinks())

        for pc_ref in self.pre_condition_state_refs:
            result.update(pc_ref._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if TYPE_CHECKING:
            self._table = odxlinks.resolve(self.table_ref, Table)
        else:
            self._table = odxlinks.resolve(self.table_ref)

        if self.dop_ref is not None:
            self._dop = odxlinks.resolve(self.dop_ref)
            if not isinstance(self._dop, (DataObjectProperty, DtcDop)):
                odxraise("The DOP-REF of TABLE-ROWs must reference a simple DOP!")
        if self.structure_ref is not None:
            self._structure = odxlinks.resolve(self.structure_ref, Structure)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

        if self.audience is not None:
            self.audience._resolve_odxlinks(odxlinks)

        self._functional_classes = NamedItemList(
            [odxlinks.resolve(fc_ref, FunctionalClass) for fc_ref in self.functional_class_refs])

        for st_ref in self.state_transition_refs:
            st_ref._resolve_odxlinks(odxlinks)

        for pc_ref in self.pre_condition_state_refs:
            pc_ref._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        # convert the raw key into the proper internal
        # representation. note that we cannot do this earlier because
        # the table's ODXLINKs must be resolved and the order of
        # ODXLINK resolution between tables and table-rows is
        # undefined.
        key_dop = self.table.key_dop
        self._key: AtomicOdxType
        if key_dop is None:
            # if the table does not define a DOP for the keys: though
            # luck, expose the raw key string. This is probably a gap
            # in the ODX specification because table-rows must exhibit
            # a "KEY" sub-tag, while the KEY-DOP-REF is optional for
            # tables (and non-existant for table rows...)
            self._key = self.key_raw
        else:
            self._key = key_dop.physical_type.base_data_type.from_string(self.key_raw)

        ddd_spec = odxrequire(context.diag_layer).diag_data_dictionary_spec

        if self.structure_snref is not None:
            self._structure = resolve_snref(self.structure_snref, ddd_spec.structures, Structure)
        if self.dop_snref is not None:
            self._dop = resolve_snref(self.dop_snref, ddd_spec.data_object_props)
            if not isinstance(self._dop, (DataObjectProperty, DtcDop)):
                odxraise("The DOP-SNREF of TABLE-ROWs must reference a simple DOP!")

        if self.audience is not None:
            self.audience._resolve_snrefs(context)

        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)

        for st_ref in self.state_transition_refs:
            st_ref._resolve_snrefs(context)

        for pc_ref in self.pre_condition_state_refs:
            pc_ref._resolve_snrefs(context)

    def __reduce__(self) -> tuple[Any, ...]:
        """This ensures that the object can be correctly reconstructed during unpickling."""
        state = self.__dict__.copy()
        return self.__class__, tuple([getattr(self, x.name) for x in fields(self)]), state



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
import argparse
import asyncio
import sys
from typing import Any

import can

import odxtools_porsche.isotp_state_machine as ism
import odxtools_porsche.uds as uds
from odxtools.diaglayers.protocol import Protocol
from odxtools.exceptions import DecodeError
from odxtools.isotp_state_machine import IsoTpStateMachine
from odxtools.response import Response, ResponseType

from . import _parser_utils
from ._parser_utils import SubparsersList

# name of the tool
_odxtools_tool_name_ = "snoop"

odx_diag_layer = None
last_request = None

ecu_rx_id = None
ecu_tx_id = None


def handle_telegram(telegram_id: int, payload: bytes) -> None:
    global odx_diag_layer
    global last_request

    assert odx_diag_layer is not None

    if telegram_id == ecu_tx_id:
        if uds.is_response_pending(payload):
            print(f" ... (response pending)")
            return

        decoded_message = None
        if last_request is not None:
            try:
                decoded_message = odx_diag_layer.decode_response(payload, last_request)
            except DecodeError:
                pass

        if decoded_message is not None:
            for i, resp in enumerate(decoded_message):
                params = resp.coding_object.parameters
                dec_str = ""
                if len(decoded_message) > 1:
                    dec_str = f" (decoding {i+1})"

                rt_str = "unknown"
                if isinstance(resp.coding_object, Response):
                    if resp.coding_object.response_type == ResponseType.POSITIVE:
                        rt_str = "positive"
                    elif resp.coding_object.response_type in (ResponseType.NEGATIVE,
                                                              ResponseType.GLOBAL_NEGATIVE):
                        rt_str = "negative"

                settable_params = []
                for param_name, param_val in resp.param_dict.items():
                    param = [x for x in params if x.short_name == param_name][0]
                    if not param.is_settable:
                        continue
                    settable_params.append((param_name, param_val))

                if settable_params:
                    print(f" {rt_str} response{dec_str} {resp.coding_object.short_name}:")
                    for param_name, param_val in settable_params:
                        print(f"      {param_name} = {repr(param_val)}")
                else:
                    print(f" {rt_str} response{dec_str} {resp.coding_object.short_name}")
        else:
            print(f" unrecognized response of {len(payload)} bytes length: "
                  f"0x{payload.hex()}")

        return

    decoded_message = None
    if odx_diag_layer is not None:
        try:
            decoded_message = odx_diag_layer.decode(payload)[0]
            last_request = payload
        except DecodeError:
            last_request = None

    if decoded_message is not None:
        print(f"request {decoded_message.coding_object.short_name}:")
        params = decoded_message.coding_object.parameters
        for param_name, param_val in decoded_message.param_dict.items():
            param = [x for x in params if x.short_name == param_name][0]
            if not param.is_settable:
                continue
            print(f"  {param_name} = {repr(param_val)}")

    else:
        print(f"Tester: "
              f"{payload.hex()} "
              f"({payload!r}, {len(payload)} bytes)")


def init_verbose_state_machine(BaseClass: type[IsoTpStateMachine], *args: Any,
                               **kwargs: Any) -> IsoTpStateMachine:

    class InformativeIsoTpDecoder(BaseClass):  # type: ignore[valid-type, misc]

        def on_sequence_error(self, telegram_idx: int, expected_idx: int, rx_idx: int) -> None:
            rx_can_id = self.can_rx_id(telegram_idx)
            print(f"Sequence error for ID 0x{rx_can_id:x}: "
                  f"Received sequence number {rx_idx} but expected {expected_idx}")

            super().on_sequence_error(telegram_idx, expected_idx, rx_idx)

        def on_frame_type_error(self, telegram_idx: int, frame_type: int) -> None:
            rx_can_id = self.can_rx_id(telegram_idx)

            print(f"Invalid ISO-TP frame for CAN ID 0x{rx_can_id:x}: {frame_type}")

    return InformativeIsoTpDecoder(*args, **kwargs)


async def active_main(args: argparse.Namespace) -> None:
    global ecu_rx_id, ecu_tx_id

    can_bus = can.Bus(channel=args.channel, bustype="socketcan")

    ecu_rx_id = int(args.rx, 0)
    ecu_tx_id = int(args.tx, 0)
    isotp_decoder = init_verbose_state_machine(
        BaseClass=ism.IsoTpActiveDecoder,
        can_bus=can_bus,
        can_rx_ids=ecu_rx_id,
        can_tx_ids=ecu_tx_id,
        padding_size=8,
    )

    print(f"Reacting to messages on channel {args.channel}")
    async for telegram_id, payload in isotp_decoder.read_telegrams(can_bus):
        handle_telegram(telegram_id, payload)


async def passive_main(args: argparse.Namespace) -> None:
    global ecu_rx_id, ecu_tx_id

    ecu_rx_id = int(args.rx, 0)
    ecu_tx_id = int(args.tx, 0)

    isotp_decoder = init_verbose_state_machine(
        BaseClass=ism.IsoTpStateMachine, can_rx_ids=[ecu_rx_id, ecu_tx_id])

    if args.channel:
        # decode a "real" bus
        can_bus = can.Bus(channel=args.channel, bustype="socketcan")

        print(f"Decoding messages on channel {args.channel}")
        async for telegram_id, payload in isotp_decoder.read_telegrams(can_bus):
            handle_telegram(telegram_id, payload)
    else:
        # decode data from stdin
        async for telegram_id, payload in isotp_decoder.read_telegrams(sys.stdin):
            handle_telegram(telegram_id, payload)


def add_cli_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--active",
        "-a",
        metavar="active_mode",
        default=False,
        action="store_const",
        const=True,
        required=False,
        help="Active mode, sends flow control messages to receive ISO-TP telegrams successfully",
    )
    parser.add_argument(
        "--channel",
        "-c",
        default=None,
        help="CAN interface name to be used (required in active mode)",
    )
    parser.add_argument(
        "--rx",
        "-r",
        default=None,
        required=False,
        help="CAN ID in which the ECU listens for diagnostic messages",
    )
    parser.add_argument(
        "--tx",
        "-t",
        default=None,
        required=False,
        help="CAN ID in which the ECU sends replys to diagnostic messages  (required in active mode)",
    )
    parser.add_argument(
        "--variant",
        "-v",
        default=None,
        required=False,
        help="Name of the ECU variant which the decode process ought to be based on",
    )
    parser.add_argument(
        "--protocol",
        "-p",
        default=None,
        required=False,
        help="Name of the protocol used for decoding",
    )
    _parser_utils.add_pdx_argument(parser)


def add_subparser(subparsers: SubparsersList) -> None:
    parser = subparsers.add_parser(
        "snoop",
        description="Live decoding of a diagnostic session.",
        help="Live decoding of a diagnostic session.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    add_cli_arguments(parser)


def run(args: argparse.Namespace) -> None:
    global odx_diag_layer
    odx_database = _parser_utils.load_file(args)

    if (odx_database is not None or args.variant is not None) and (odx_database is None or
                                                                   args.variant is None):
        print("The database and variant must always either be "
              "both specified or not at all")
        sys.exit(1)

    odx_diag_layer = None
    if odx_database is not None:
        odx_diag_layer = odx_database.diag_layers.get(args.variant)

        if odx_diag_layer is None:
            print(f"Variant '{args.variant}' does not exist. Available variants:")
            for dl in odx_database.diag_layers:
                desc = "" if dl.description is None else f": {dl.description}"
                print(f"  {dl.short_name}{desc}")
            sys.exit(1)

    protocol_name = args.protocol
    if odx_diag_layer is not None and protocol_name is not None:
        protocols: list[Protocol] | None = getattr(odx_diag_layer, "protocols", None)

        if protocols is None:
            print(f"ECU variant {odx_diag_layer.short_name} is of type "
                  f"{odx_diag_layer.variant_type.value} and thus does not "
                  f"feature any protocols")
            sys.exit(1)

        protocol_names = [x.short_name for x in protocols]
        if protocol_name not in protocol_names:
            print(f"ECU variant {odx_diag_layer.short_name} does not support "
                  f"a protocol named '{protocol_name}'. Supported protocols are:")
            for x in protocols:
                desc = "" if x.description is None else f": {x.description}"
                print(f"  {x.short_name}{desc}")
            sys.exit(1)

    # if no can IDs have been explicitly specified, take them from the DL
    if args.rx is None and odx_diag_layer:
        get_can_rx_id = getattr(odx_diag_layer, "get_can_receive_id", None)
        if get_can_rx_id is None:
            print(f"ECU variant {odx_diag_layer.short_name} is of type "
                  f"{odx_diag_layer.variant_type.value} and thus does not "
                  f"provide any communication parameters")
            sys.exit(1)

        args.rx = str(get_can_rx_id(protocol=protocol_name))

    if args.tx is None and odx_diag_layer:
        get_can_tx_id = getattr(odx_diag_layer, "get_can_send_id", None)
        if get_can_tx_id is None:
            print(f"ECU variant {odx_diag_layer.short_name} is of type "
                  f"{odx_diag_layer.variant_type.value} and thus does not "
                  f"provide any communication parameters")
            sys.exit(1)

        args.tx = str(get_can_tx_id(protocol=protocol_name))

    if args.rx is None:
        print(f"Could not determine a CAN receive ID.")
        sys.exit(1)

    if args.active:
        asyncio.run(active_main(args))
    else:
        asyncio.run(passive_main(args))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Decode ISO-TP communication over a can bus")

    add_cli_arguments(parser)

    args = parser.parse_args()  # deals with the help message handling

    run(args)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from typing_extensions import override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..odxdoccontext import OdxDocContext
from ..odxtypes import ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType


@dataclass(kw_only=True)
class DynamicParameter(Parameter):

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "DYNAMIC"

    @property
    @override
    def is_required(self) -> bool:
        raise NotImplementedError(".is_required for a DynamicParameter")

    @property
    @override
    def is_settable(self) -> bool:
        raise NotImplementedError(".is_settable for a DynamicParameter")

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DynamicParameter":

        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        return DynamicParameter(**kwargs)

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        raise NotImplementedError("Encoding DynamicParameter is not implemented yet.")

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        raise NotImplementedError("Decoding DynamicParameter is not implemented yet.")



# SPDX-License-Identifier: MIT
from collections.abc import Generator
from copy import copy
from enum import Enum

from .diaglayers.basevariant import BaseVariant
from .diaglayers.ecuvariant import EcuVariant
from .exceptions import DecodeError, odxraise
from .matchingparameter import MatchingParameter
from .response import Response


class VariantMatcher:
    """VariantMatcher implements the matching algorithm of ECU and
    base variants according to their ECU-VARIANT-PATTERNs or
    BASE-VARIANT-PATTERNs according to ISO 22901-1.

        Usage (example):

    ```python

    # initialize the matcher with a list of base or ECU variants
    matcher = VariantMatcher(candidates=[...], use_cache=use_cache)

    # run the request loop to obtain responses for every request
    for use_physical_addressing, encoded_request in matcher.request_loop():
        if use_physical_addressing:
            resp = send_to_ecu_using_physical_addressing(encoded_request)
        else:
            resp = send_to_ecu_using_functional_addressing(encoded_request)
        matcher.evaluate(resp)

    # result
    if matcher.has_match():
        if isinstance(matcher.matching_variant, BaseVariant):
            print(f"Match found for base variant "
                  f"{matcher.matching_variant.short_name}")
        elif isinstance(matcher.matching_variant, EcuVariant):
            print(f"Match found for ECU variant "
                  f"{matcher.matching_variant.short_name}")
        else:
            print(f"Match found for unknown diag layer type "
                  f"{type(matcher.matching_variant).__name__}")
    else:
        print("No matching base- or ECU variant found")
    ```

    TODO: Note that only patterns that exclusivly reference diagnostic
    services (i.e., no single-ECU jobs) in their matching parameters
    are currently supported.
    """

    class State(Enum):
        PENDING = 0
        NO_MATCH = 1
        MATCH = 2

    def __init__(self,
                 variant_candidates: list[EcuVariant] | list[BaseVariant],
                 use_cache: bool = True):

        self.variant_candidates = variant_candidates
        self.use_cache = use_cache
        self.req_resp_cache: dict[bytes, bytes] = {}
        self._recent_ident_response: bytes | None = None

        self._state = VariantMatcher.State.PENDING
        self._matching_variant: EcuVariant | BaseVariant | None = None

    def request_loop(self) -> Generator[tuple[bool, bytes], None, None]:
        """The request loop yielding tuples of byte sequences of
        requests and the whether physical addressing ought to be used
        to send them

        Each of these requests needs to be send to the ECU to be
        identified using the specified addressing scheme. The response
        of the ECU is then required to be passed to the matcher using
        the `evaluate()` method.
        """
        from .basevariantpattern import BaseVariantPattern
        from .ecuvariantpattern import EcuVariantPattern
        from .matchingbasevariantparameter import MatchingBaseVariantParameter

        if not self.is_pending():
            return

        self._matching_variant = None
        for variant in self.variant_candidates:
            variant_patterns: list[EcuVariantPattern] | list[BaseVariantPattern]
            if isinstance(variant, EcuVariant):
                variant_patterns = variant.ecu_variant_patterns
            elif isinstance(variant, BaseVariant):
                if variant.base_variant_pattern is None:
                    variant_patterns = []
                else:
                    variant_patterns = [variant.base_variant_pattern]
            else:
                odxraise(f"Only EcuVariant and BaseVariant are supported "
                         f"for pattern matching, not {type(self).__name__}")
                self._state = VariantMatcher.State.NO_MATCH
                return

            any_pattern_matches = False
            for pattern in variant_patterns:
                all_params_match = True
                for matching_param in pattern.get_matching_parameters():
                    req_bytes = matching_param.get_ident_service(variant).encode_request()

                    if self.use_cache and req_bytes in self.req_resp_cache:
                        resp_values = copy(self.req_resp_cache[req_bytes])
                    else:
                        if isinstance(matching_param, MatchingBaseVariantParameter):
                            yield matching_param.use_physical_addressing, req_bytes
                        else:
                            yield True, req_bytes
                        resp_values = self._get_ident_response()
                        self._update_cache(req_bytes, copy(resp_values))

                    cur_response_matches = self._ident_response_matches(
                        variant, matching_param, resp_values)
                    all_params_match = all_params_match and cur_response_matches
                    if not all_params_match:
                        break

                if all_params_match:
                    any_pattern_matches = True
                    break

            if any_pattern_matches:
                self._state = VariantMatcher.State.MATCH
                self._matching_variant = variant
                break

        if self.is_pending():
            # no pattern has matched for any ecu variant
            self._state = VariantMatcher.State.NO_MATCH

    def evaluate(self, resp_bytes: bytes) -> None:
        """Update the matcher with the response to a requst.

        Warning: Use this method EXACTLY once within the loop body of the request loop.
        """
        self._recent_ident_response = bytes(resp_bytes)

    def is_pending(self) -> bool:
        """True iff request loop has not yet been run."""
        return self._state == VariantMatcher.State.PENDING

    def has_match(self) -> bool:
        """Returns true iff the non-pending matcher found a matching ecu variant.

        Raises a runtime error if the matcher is pending.
        """
        if self.is_pending():
            raise RuntimeError(
                "EcuVariantMatcher is pending. Run the request_loop to determine the active ecu variant."
            )
        return self._state == VariantMatcher.State.MATCH

    @property
    def matching_variant(self) -> EcuVariant | BaseVariant | None:
        """Returns the matched, i.e., active ecu variant if such a variant has been found."""
        return self._matching_variant

    def _ident_response_matches(
        self,
        variant: EcuVariant | BaseVariant,
        matching_param: MatchingParameter,
        response_bytes: bytes,
    ) -> bool:
        """Decode a binary response and extract the identification string according
        to the snref or snpathref of the matching_param.
        """
        service = matching_param.get_ident_service(variant)

        # ISO 22901 requires that snref or snpathref is resolvable in
        # at least one POS-RESPONSE or NEG-RESPONSE
        all_responses: list[Response] = []
        all_responses.extend(service.positive_responses)
        all_responses.extend(service.negative_responses)
        all_responses.extend(variant.global_negative_responses)

        for cur_response in all_responses:
            try:
                decoded_vals = cur_response.decode(response_bytes)
            except DecodeError:
                # the current response object could not decode the received
                # data. Ignore it.
                continue

            if not matching_param.matches(decoded_vals):
                # This particular response object does not match the
                # expected value of the specified matching
                # parameter. Ignore it.
                continue

            return True

        return False

    def _update_cache(self, req_bytes: bytes, resp_bytes: bytes) -> None:
        if self.use_cache:
            self.req_resp_cache[req_bytes] = resp_bytes

    def _get_ident_response(self) -> bytes:
        if not self._recent_ident_response:
            raise RuntimeError(
                "No response available. Did you forget to call 'evaluate()' in a loop?")
        return self._recent_ident_response



# SPDX-License-Identifier: MIT
from collections.abc import Sequence
from dataclasses import dataclass
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .encodestate import EncodeState
from .exceptions import EncodeError, odxassert, odxraise
from .field import Field
from .odxdoccontext import OdxDocContext
from .odxtypes import ParameterValue
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class EndOfPduField(Field):
    """End of PDU fields are structures that are repeated until the end of the PDU"""
    max_number_of_items: int | None = None
    min_number_of_items: int | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "EndOfPduField":
        kwargs = dataclass_fields_asdict(Field.from_et(et_element, context))

        if (max_n_str := et_element.findtext("MAX-NUMBER-OF-ITEMS")) is not None:
            max_number_of_items = int(max_n_str)
        else:
            max_number_of_items = None
        if (min_n_str := et_element.findtext("MIN-NUMBER-OF-ITEMS")) is not None:
            min_number_of_items = int(min_n_str)
        else:
            min_number_of_items = None

        return EndOfPduField(
            max_number_of_items=max_number_of_items,
            min_number_of_items=min_number_of_items,
            **kwargs)

    @override
    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        odxassert(not encode_state.cursor_bit_position,
                  "No bit position can be specified for end-of-pdu fields!")
        odxassert(encode_state.is_end_of_pdu,
                  "End-of-pdu fields can only be located at the end of PDUs!")

        if not isinstance(physical_value, Sequence):
            odxraise(
                f"Invalid type {type(physical_value).__name__} of physical "
                f"value for end-of-pdu field, expected a list", EncodeError)
            return

        orig_is_end_of_pdu = encode_state.is_end_of_pdu
        encode_state.is_end_of_pdu = False

        for i, value in enumerate(physical_value):
            if i == len(physical_value) - 1:
                encode_state.is_end_of_pdu = orig_is_end_of_pdu

            self.structure.encode_into_pdu(value, encode_state)

        encode_state.is_end_of_pdu = orig_is_end_of_pdu

    @override
    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        odxassert(not decode_state.cursor_bit_position,
                  "No bit position can be specified for end-of-pdu fields!")

        orig_origin = decode_state.origin_byte_position
        decode_state.origin_byte_position = decode_state.cursor_byte_position

        result: list[ParameterValue] = []
        while decode_state.cursor_byte_position < len(decode_state.coded_message):
            # ATTENTION: the ODX specification is very misleading
            # here: it says that the item is repeated until the end of
            # the PDU, but it means that DOP of the items that are
            # repeated are identical, not their values
            result.append(self.structure.decode_from_pdu(decode_state))

        decode_state.origin_byte_position = orig_origin

        return result



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
import argparse

import odxtools_porsche
from examples import somersaultecu

argparser = argparse.ArgumentParser(
    description="\n".join([
        "Creates a simple sample PDX file for a 'somersault' ECU from scratch.",
        "",
        "This is primarily intended to be a demonstration of how to define",
        "automotive diagnostic services in python and making them interact",
        "with existing tooling.",
    ]),
    formatter_class=argparse.RawTextHelpFormatter,
)

argparser.add_argument(
    "output_pdx_file",
    metavar="OUTPUT_PDX_FILE",
    help="Path to the where the resulting .pdx file is written",
)

args = argparser.parse_args()

odxtools.write_pdx_file(args.output_pdx_file, somersaultecu.database)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import NamedElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext
from .structure import Structure
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class MultiplexerDefaultCase(NamedElement):
    """This class represents a Default Case, which is selected when there are no cases defined in the Multiplexer."""
    structure_ref: OdxLinkRef | None = None
    structure_snref: str | None = None

    @property
    def structure(self) -> Structure | None:
        return self._structure

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "MultiplexerDefaultCase":
        """Reads a default case for a multiplexer."""
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        structure_ref = OdxLinkRef.from_et(et_element.find("STRUCTURE-REF"), context)
        structure_snref = None
        if (structure_snref_elem := et_element.find("STRUCTURE-SNREF")) is not None:
            structure_snref = odxrequire(structure_snref_elem.get("SHORT-NAME"))

        return MultiplexerDefaultCase(
            structure_ref=structure_ref, structure_snref=structure_snref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._structure = None
        if self.structure_ref is not None:
            self._structure = odxlinks.resolve(self.structure_ref, Structure)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.structure_snref:
            ddds = odxrequire(context.diag_layer).diag_data_dictionary_spec
            self._structure = resolve_snref(self.structure_snref, ddds.structures, Structure)



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
import argparse

import odxtools_porsche

argparser = argparse.ArgumentParser(
    description="\n".join([
        "Read in a PDX file and write back the resulting database object.",
        "",
        "This is primarily intended to be a demonstration of how to write",
        "PDX files using odxtools, but it can also be used to strip",
        "'unnecessary' auxiliary data from the input PDX file.",
    ]),
    formatter_class=argparse.RawTextHelpFormatter,
)

argparser.add_argument(
    "input_pdx_file", metavar="INPUT_PDX_FILE", help="Path to the input .pdx file")
argparser.add_argument(
    "output_pdx_file",
    metavar="OUTPUT_PDX_FILE",
    help="Path to the where the resulting .pdx file is written",
)

args = argparser.parse_args()

in_file_name = args.input_pdx_file
out_file_name = args.output_pdx_file

print(f"Loading input file '{in_file_name}'...", end="", flush=True)
db = odxtools.load_pdx_file(in_file_name)
print(" done")

print(f"Writing output file '{out_file_name}'...", end="", flush=True)
odxtools.write_pdx_file(out_file_name, db)
print(" done")



from dataclasses import dataclass
from xml.etree import ElementTree

from .odxdoccontext import OdxDocContext


@dataclass(kw_only=True)
class Text:
    text: str
    text_identifier: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Text":
        # Extract the contents of the tag as a string.
        raw_string = et_element.text or ""
        for e in et_element:
            raw_string += ElementTree.tostring(e, encoding="unicode")

        # remove white spaces at the beginning and at the end of all
        # extracted lines
        stripped_lines = [x.strip() for x in raw_string.split("\n")]

        text = "\n".join(stripped_lines).strip()

        text_identifier = et_element.get("TI")

        return Text(text=text, text_identifier=text_identifier)

    @staticmethod
    def from_string(text: str) -> "Text":
        return Text(text=text, text_identifier=None)

    def __str__(self) -> str:
        return self.text



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import NamedElement
from .environmentdata import EnvironmentData
from .environmentdatadescription import EnvironmentDataDescription
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class EnvDataConnector(NamedElement):
    env_data_desc_ref: OdxLinkRef
    env_data_snref: str

    @property
    def env_data_desc(self) -> EnvironmentDataDescription:
        return self._env_data_desc

    @property
    def env_data(self) -> EnvironmentData:
        return self._env_data

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "EnvDataConnector":
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        env_data_desc_ref = odxrequire(
            OdxLinkRef.from_et(et_element.find("ENV-DATA-DESC-REF"), context))
        env_data_snref_el = odxrequire(et_element.find("ENV-DATA-SNREF"))
        env_data_snref = odxrequire(env_data_snref_el.get("SHORT-NAME"))

        return EnvDataConnector(
            env_data_desc_ref=env_data_desc_ref, env_data_snref=env_data_snref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._env_data_desc = odxlinks.resolve(self.env_data_desc_ref, EnvironmentDataDescription)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        self._env_data = resolve_snref(self.env_data_snref, self._env_data_desc.env_datas,
                                       EnvironmentData)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, cast
from xml.etree import ElementTree

from typing_extensions import override

from ..decodestate import DecodeState
from ..encodestate import EncodeState
from ..exceptions import odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkRef
from ..odxtypes import ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType
from .rowfragment import RowFragment

if TYPE_CHECKING:
    from ..tablerow import TableRow


@dataclass(kw_only=True)
class TableEntryParameter(Parameter):
    target: RowFragment
    table_row_ref: OdxLinkRef

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "TABLE-ENTRY"

    @property
    @override
    def is_required(self) -> bool:
        raise NotImplementedError("TableEntryParameter.is_required is not implemented yet.")

    @property
    @override
    def is_settable(self) -> bool:
        raise NotImplementedError("TableEntryParameter.is_settable is not implemented yet.")

    @property
    def table_row(self) -> "TableRow":
        return self._table_row

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "TableEntryParameter":
        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        target_str = odxrequire(et_element.findtext("TARGET"))
        try:
            target = RowFragment(target_str)
        except ValueError:
            odxraise(f"Encountered unknown target '{target_str}'")
            target = cast(RowFragment, None)
        table_row_ref = odxrequire(OdxLinkRef.from_et(et_element.find("TABLE-ROW-REF"), context))

        return TableEntryParameter(target=target, table_row_ref=table_row_ref, **kwargs)

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        if TYPE_CHECKING:
            self._table_row = odxlinks.resolve(self.table_row_ref, TableRow)
        else:
            self._table_row = odxlinks.resolve(self.table_row_ref)

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        raise NotImplementedError("Encoding a TableEntryParameter is not implemented yet.")

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        raise NotImplementedError("Decoding a TableEntryParameter is not implemented yet.")



# SPDX-License-Identifier: MIT
import typing
from dataclasses import dataclass, field
from typing import Any, runtime_checkable
from xml.etree import ElementTree

from .admindata import AdminData
from .commrelation import CommRelation
from .element import IdentifiableElement
from .exceptions import odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .odxtypes import odxstr_to_bool
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .swvariable import SwVariable
from .table import Table
from .tablerow import TableRow
from .utils import dataclass_fields_asdict
from .variablegroup import VariableGroup


@runtime_checkable
class HasDiagVariables(typing.Protocol):

    @property
    def diag_variables(self) -> "NamedItemList[DiagVariable]":
        ...


@dataclass(kw_only=True)
class DiagVariable(IdentifiableElement):
    """Representation of a diagnostic variable
    """

    admin_data: AdminData | None = None
    variable_group_ref: OdxLinkRef | None = None
    sw_variables: list[SwVariable] = field(default_factory=list)

    # a diag variable must specify either COMM-RELATIONS or a
    # reference to a table row
    comm_relations: list[CommRelation] = field(default_factory=list)

    # these are nested inside the SNREF-TO-TABLEROW tag
    table_snref: str | None = None
    table_row_snref: str | None = None

    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    is_read_before_write_raw: bool | None = None

    @property
    def table(self) -> Table | None:
        return self._table

    @property
    def table_row(self) -> TableRow | None:
        return self._table_row

    @property
    def variable_group(self) -> VariableGroup | None:
        return self._variable_group

    @property
    def is_read_before_write(self) -> bool:
        return self.is_read_before_write_raw is True

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagVariable":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)
        variable_group_ref = OdxLinkRef.from_et(et_element.find("VARIABLE-GROUP-REF"), context)
        sw_variables = NamedItemList([
            SwVariable.from_et(swv_elem, context)
            for swv_elem in et_element.iterfind("SW-VARIABLES/SW-VARIABLE")
        ])
        comm_relations = [
            CommRelation.from_et(cr_elem, context)
            for cr_elem in et_element.iterfind("COMM-RELATIONS/COMM-RELATION")
        ]

        table_snref = None
        table_row_snref = None
        if (snref_to_tablerow_elem := et_element.find("SNREF-TO-TABLEROW")) is not None:
            table_snref_elem = odxrequire(snref_to_tablerow_elem.find("TABLE-SNREF"))
            table_snref = odxrequire(table_snref_elem.attrib.get("SHORT-NAME"))

            table_row_snref_elem = odxrequire(snref_to_tablerow_elem.find("TABLE-ROW-SNREF"))
            table_row_snref = odxrequire(table_row_snref_elem.attrib.get("SHORT-NAME"))

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        is_read_before_write_raw = odxstr_to_bool(et_element.get("IS-READ-BEFORE-WRITE"))

        return DiagVariable(
            admin_data=admin_data,
            variable_group_ref=variable_group_ref,
            sw_variables=sw_variables,
            comm_relations=comm_relations,
            table_snref=table_snref,
            table_row_snref=table_row_snref,
            sdgs=sdgs,
            is_read_before_write_raw=is_read_before_write_raw,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        if self.admin_data is not None:
            result.update(self.admin_data._build_odxlinks())

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        for cr in self.comm_relations:
            result.update(cr._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_odxlinks(odxlinks)

        self._variable_group = None
        if self.variable_group_ref is not None:
            self._variable_group = odxlinks.resolve(self.variable_group_ref, VariableGroup)

        for cr in self.comm_relations:
            cr._resolve_odxlinks(odxlinks)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_snrefs(context)

        for cr in self.comm_relations:
            cr._resolve_snrefs(context)

        self._table = None
        self._table_row = None
        if self.table_snref is not None:
            ddds = odxrequire(context.diag_layer).diag_data_dictionary_spec
            self._table = resolve_snref(self.table_snref, ddds.tables, Table)
            self._table_row = resolve_snref(
                odxrequire(self.table_row_snref), self._table.table_rows, TableRow)

        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, cast
from xml.etree import ElementTree

from .addressing import Addressing
from .comparaminstance import ComparamInstance
from .diagcomm import DiagComm
from .exceptions import DecodeError, DecodeMismatch, odxassert, odxraise, odxrequire
from .message import Message
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import ParameterValue, odxstr_to_bool
from .parameters.parameter import Parameter
from .posresponsesuppressible import PosResponseSuppressible
from .request import Request
from .response import Response
from .snrefcontext import SnRefContext
from .transmode import TransMode
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DiagService(DiagComm):
    """Representation of a diagnostic service description.
    """

    comparam_refs: list[ComparamInstance] = field(default_factory=list)
    request_ref: OdxLinkRef
    pos_response_refs: list[OdxLinkRef] = field(default_factory=list)
    neg_response_refs: list[OdxLinkRef] = field(default_factory=list)
    pos_response_suppressible: PosResponseSuppressible | None = None

    is_cyclic_raw: bool | None = None
    is_multiple_raw: bool | None = None
    addressing_raw: Addressing | None = None
    transmission_mode_raw: TransMode | None = None

    @property
    def comparams(self) -> NamedItemList[ComparamInstance]:
        return self._comparams

    @property
    def request(self) -> Request | None:
        return self._request

    @property
    def positive_responses(self) -> NamedItemList[Response]:
        return self._positive_responses

    @property
    def negative_responses(self) -> NamedItemList[Response]:
        return self._negative_responses

    @property
    def is_cyclic(self) -> bool:
        return self.is_cyclic_raw is True

    @property
    def is_multiple(self) -> bool:
        return self.is_multiple_raw is True

    @property
    def addressing(self) -> Addressing:
        return self.addressing_raw or Addressing.PHYSICAL

    @property
    def transmission_mode(self) -> TransMode:
        return self.transmission_mode_raw or TransMode.SEND_AND_RECEIVE

    @property
    def free_parameters(self) -> list[Parameter]:
        """Return the list of parameters which can be freely specified by
        the user when encoding the service's request.
        """
        return self.request.free_parameters if self.request is not None else []

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagService":

        kwargs = dataclass_fields_asdict(DiagComm.from_et(et_element, context))

        comparam_refs = [
            ComparamInstance.from_et(el, context)
            for el in et_element.iterfind("COMPARAM-REFS/COMPARAM-REF")
        ]

        request_ref = odxrequire(OdxLinkRef.from_et(et_element.find("REQUEST-REF"), context))

        pos_response_refs = [
            odxrequire(OdxLinkRef.from_et(el, context))
            for el in et_element.iterfind("POS-RESPONSE-REFS/POS-RESPONSE-REF")
        ]

        neg_response_refs = [
            odxrequire(OdxLinkRef.from_et(el, context))
            for el in et_element.iterfind("NEG-RESPONSE-REFS/NEG-RESPONSE-REF")
        ]

        pos_response_suppressible = None
        if (prs_elem := et_element.find("POS-RESPONSE-SUPPRESSABLE")) is not None:
            pos_response_suppressible = PosResponseSuppressible.from_et(prs_elem, context)

        is_cyclic_raw = odxstr_to_bool(et_element.get("IS-CYCLIC"))
        is_multiple_raw = odxstr_to_bool(et_element.get("IS-MULTIPLE"))

        addressing_raw: Addressing | None = None
        if (addressing_raw_str := et_element.get("ADDRESSING")) is not None:
            try:
                addressing_raw = Addressing(addressing_raw_str)
            except ValueError:
                addressing_raw = cast(Addressing, None)
                odxraise(f"Encountered unknown addressing type '{addressing_raw_str}'")

        transmission_mode_raw: TransMode | None = None
        if (transmission_mode_raw_str := et_element.get("TRANSMISSION-MODE")) is not None:
            try:
                transmission_mode_raw = TransMode(transmission_mode_raw_str)
            except ValueError:
                transmission_mode_raw = cast(TransMode, None)
                odxraise(f"Encountered unknown transmission mode '{transmission_mode_raw_str}'")

        return DiagService(
            comparam_refs=comparam_refs,
            request_ref=request_ref,
            pos_response_refs=pos_response_refs,
            neg_response_refs=neg_response_refs,
            pos_response_suppressible=pos_response_suppressible,
            is_cyclic_raw=is_cyclic_raw,
            is_multiple_raw=is_multiple_raw,
            addressing_raw=addressing_raw,
            transmission_mode_raw=transmission_mode_raw,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for cpr in self.comparam_refs:
            result.update(cpr._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for cpr in self.comparam_refs:
            cpr._resolve_odxlinks(odxlinks)

        self._request = odxlinks.resolve(self.request_ref, Request)

        self._positive_responses = NamedItemList[Response](
            [odxlinks.resolve(x, Response) for x in self.pos_response_refs])

        self._negative_responses = NamedItemList[Response](
            [odxlinks.resolve(x, Response) for x in self.neg_response_refs])

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        context.diag_service = self

        super()._resolve_snrefs(context)

        for cpr in self.comparam_refs:
            cpr._resolve_snrefs(context)

        # The named item list of communication parameters is created
        # here because ComparamInstance.short_name is only valid after
        # reference resolution
        self._comparams = NamedItemList(self.comparam_refs)

        context.diag_service = None

    def print_free_parameters_info(self) -> None:
        """Return a human readable description of the service's
        request's free parameters.
        """
        if self.request is None:
            return

        self.request.print_free_parameters_info()

    def decode_message(self, raw_message: bytes) -> Message:
        request_prefix = b''
        candidate_coding_objects: list[Request | Response] = [
            *self.positive_responses, *self.negative_responses
        ]
        if self.request is not None:
            request_prefix = self.request.coded_const_prefix()
            candidate_coding_objects.append(self.request)

        coding_objects: list[Request | Response] = []
        for candidate_coding_object in candidate_coding_objects:
            prefix = candidate_coding_object.coded_const_prefix(request_prefix=request_prefix)
            if len(raw_message) >= len(prefix) and prefix == raw_message[:len(prefix)]:
                coding_objects.append(candidate_coding_object)

        result_list: list[Message] = []
        for coding_object in coding_objects:
            try:
                result_list.append(
                    Message(
                        coded_message=raw_message,
                        service=self,
                        coding_object=coding_object,
                        param_dict=coding_object.decode(raw_message)))
            except DecodeMismatch:
                # An NRC-CONST or environment data parameter
                # encountered a non-matching value -> coding object
                # does not apply
                pass

        if len(result_list) < 1:
            odxraise(f"The service {self.short_name} cannot decode the message {raw_message.hex()}",
                     DecodeError)
            return Message(
                coded_message=raw_message, service=self, coding_object=None, param_dict={})
        elif len(result_list) > 1:
            odxraise(
                f"The service {self.short_name} cannot uniquely decode the message {raw_message.hex()}",
                DecodeError)

        return result_list[0]

    def encode_request(self, **kwargs: ParameterValue) -> bytes:
        """Prepare an array of bytes ready to be send over the wire
        for the request of this service.
        """
        # make sure that all parameters which are required for
        # encoding are specified (parameters which have a default are
        # optional)
        if self.request is None:
            return b''

        missing_params = {x.short_name
                          for x in self.request.required_parameters}.difference(kwargs.keys())
        odxassert(
            len(missing_params) == 0, f"The parameters {missing_params} are required but missing!")

        # make sure that no unknown parameters are specified
        rq_all_param_names = {x.short_name for x in self.request.parameters}
        odxassert(
            set(kwargs.keys()).issubset(rq_all_param_names),
            f"Unknown parameters specified for encoding: {kwargs.keys()}, "
            f"known parameters are: {rq_all_param_names}")
        return self.request.encode(**kwargs)

    def encode_positive_response(self,
                                 coded_request: bytes,
                                 response_index: int = 0,
                                 **kwargs: ParameterValue) -> bytes:
        # TODO: Should the user decide the positive response or what are the differences?
        return self.positive_responses[response_index].encode(coded_request, **kwargs)

    def encode_negative_response(self,
                                 coded_request: bytes,
                                 response_index: int = 0,
                                 **kwargs: ParameterValue) -> bytes:
        return self.negative_responses[response_index].encode(coded_request, **kwargs)

    def __call__(self, **kwargs: ParameterValue) -> bytes:
        """Encode a request."""
        return self.encode_request(**kwargs)



# SPDX-License-Identifier: MIT
import unittest

from odxtools.odxtypes import DataType


class TestDataType(unittest.TestCase):

    def test_from_string(self) -> None:
        self.assertTrue(DataType.A_UINT32 == DataType(DataType.A_UINT32))

    def test_make_from(self) -> None:
        self.assertTrue(DataType.A_UINT32.make_from("12") == 12)
        self.assertTrue(DataType.A_FLOAT64.make_from("3.14") == 3.14)
        self.assertTrue(DataType.A_INT32.make_from(3.14) == 3)

        self.assertTrue(DataType.A_BYTEFIELD.make_from("12") == b"\x12")

        self.assertTrue(DataType.A_UINT32.from_string("12") == 12)
        self.assertTrue(DataType.A_BYTEFIELD.from_string("12") == b"\x12")

    def test_python_types(self) -> None:
        self.assertTrue(DataType.A_UINT32.python_type is int)
        self.assertTrue(DataType.A_INT32.python_type is int)
        self.assertTrue(DataType.A_FLOAT32.python_type is float)
        self.assertTrue(DataType.A_FLOAT64.python_type is float)
        self.assertTrue(DataType.A_BYTEFIELD.python_type is bytearray)
        self.assertTrue(DataType.A_UNICODE2STRING.python_type is str)
        self.assertTrue(DataType.A_UTF8STRING.python_type is str)
        self.assertTrue(DataType.A_ASCIISTRING.python_type is str)

    def test_isinstance(self) -> None:
        self.assertTrue(DataType.A_ASCIISTRING.isinstance("123"))
        self.assertTrue(DataType.A_BYTEFIELD.isinstance(bytes([0x12, 0x34])))
        self.assertTrue(DataType.A_UINT32.isinstance(123))
        self.assertTrue(DataType.A_FLOAT32.isinstance(123.456))
        # We allow integers as float values
        self.assertTrue(DataType.A_FLOAT32.isinstance(123))

        self.assertFalse(DataType.A_UINT32.isinstance(bytes([0x12])))
        self.assertFalse(DataType.A_BYTEFIELD.isinstance([0x12]))
        self.assertFalse(DataType.A_BYTEFIELD.isinstance(0x12))


if __name__ == "__main__":
    unittest.main()
